# -*- coding: utf-8 -*-
"""
string normalizer enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class NormalizerEnum(CoreEnum):
    """
    normalizer enum.
    """

    SPACE = 'space'
    DUPLICATE_SPACE = 'duplicate_space'
    LOWERCASE = 'lowercase'
    UPPERCASE = 'uppercase'
    TITLE_CASE = 'title_case'
    FILTER = 'filter'
    PERSIAN_SIGN = 'persian_sign'
    LATIN_SIGN = 'latin_sign'
    PERSIAN_NUMBER = 'persian_number'
    ARABIC_NUMBER = 'arabic_number'
    PERSIAN_LETTER = 'persian_letter'
    LATIN_LETTER = 'latin_letter'



# -*- coding: utf-8 -*-
"""
response wrappers text module.
"""

from pyrin.processor.mimetype.enumerations import MIMETypeEnum
from pyrin.processor.response.wrappers.base import CoreResponse


class TextResponse(CoreResponse):
    """
    text response class.
    """

    default_mimetype = MIMETypeEnum.TEXT

    def __init__(self, response=None, status=None,
                 headers=None, direct_passthrough=False,
                 **options):
        """
        initializes an instance of TextResponse.

        :param str response: an object containing the response value.

        :param str | int status: a string with a status or an integer
                                 with the status code.

        :param list | Headers headers: a list of headers or a
                                       `datastructures.Headers` object.

        :param bool direct_passthrough: if set to True the `iter_encoded` method is not
                                        called before iteration which makes it
                                        possible to pass special iterators through
                                        unchanged.

        :keyword dict original_data: a dict containing the original
                                     data of response before encoding.
                                     this value will be used in logging
                                     to mask critical values.
        """

        super().__init__(response, status, headers,
                         mimetype=self.default_mimetype,
                         direct_passthrough=direct_passthrough,
                         **options)



# -*- coding: utf-8 -*-
"""
template exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class TemplateManagerException(CoreException):
    """
    template manager exception.
    """
    pass


class TemplateManagerBusinessException(CoreBusinessException, TemplateManagerException):
    """
    template manager business exception.
    """
    pass


class TemplateHandlerNameRequiredError(TemplateManagerException):
    """
    template handler name required error.
    """
    pass


class InvalidSourceDirectoryError(TemplateManagerException):
    """
    invalid source directory error.
    """
    pass


class InvalidTargetDirectoryError(TemplateManagerException):
    """
    invalid target directory error.
    """
    pass


class InvalidTemplateHandlerTypeError(TemplateManagerException):
    """
    invalid template handler type error.
    """
    pass


class DuplicatedTemplateHandlerError(TemplateManagerException):
    """
    duplicated template handler error.
    """
    pass


class TemplateHandlerNotFoundError(TemplateManagerBusinessException):
    """
    template handler not found error.
    """
    pass


class TemplateTargetDirectoryAlreadyExistedError(TemplateManagerException):
    """
    template target directory already existed error.
    """
    pass



# -*- coding: utf-8 -*-
"""
users api module.
"""

from pyrin.api.router.decorators import api
from pyrin.core.enumerations import HTTPMethodEnum

import demo.security.users.services as user_services

from demo.security.users.permissions import VIEW_USERS_LIST_PERMISSION


@api('/users', methods=HTTPMethodEnum.POST, authenticated=False)
def create(username, password, first_name, last_name, **options):
    """
    creates a new user based on given inputs.

    :param str username: username.
    :param str password: password.
    :param str first_name: first name.
    :param str last_name: last name.
    """

    user_services.create(username, password, first_name, last_name, **options)


@api('/users/info', methods=HTTPMethodEnum.GET, permissions=VIEW_USERS_LIST_PERMISSION)
def get_info(**options):
    """
    gets the current user info.

    :raises UserNotFoundError: user not found error.

    :returns: dict(int id,
                   str username,
                   datetime last_login_date,
                   str first_name,
                   str last_name,
                   bool is_active)
    :rtype: dict
    """

    return user_services.get(**options)


@api('/users', methods=HTTPMethodEnum.GET, authenticated=False)
def get_all(**options):
    """
    gets a list of all users.

    :returns: list[dict(int id,
                        str username,
                        datetime last_login_date,
                        str first_name,
                        str last_name,
                        bool is_active)]
    :rtype: list
    """

    return user_services.get_all(**options), {'Access-Control-Allow-Origin': 'HTTP://127.0.0.1:5000'}


@api('/page', authenticated=False)
def file(**options):
    return '<!DOCTYPE html><html><body><h1>The XMLHttpRequest Object</h1>' \
'<button type="button" onclick="loadDoc()">Request data</button>' \
'<p id="demo"></p><script>function loadDoc() ' \
'{var xhttp = new XMLHttpRequest();xhttp.onreadystatechange = function() ' \
'{if (this.readyState == 4 && this.status == 200) {' \
'document.getElementById("demo").innerHTML = this.responseText;}};' \
'xhttp.open("GET", "http://128.65.185.94:5000/accounts/", true);' \
'xhttp.send();}</script></body></html>'


# -*- coding: utf-8 -*-
"""
mimetype handlers json module.
"""

import re

from pyrin.processor.mimetype.decorators import mimetype_handler
from pyrin.processor.mimetype.enumerations import MIMETypeEnum
from pyrin.processor.mimetype.handlers.base import MIMETypeHandlerBase


@mimetype_handler()
class JSONMIMETypeHandler(MIMETypeHandlerBase):
    """
    json mimetype handler class.
    """

    def __init__(self, **options):
        """
        creates an instance of JSONMIMETypeHandler.
        """

        super().__init__(**options)

    def _mimetype(self, value, **options):
        """
        gets the mimetype of given value.

        returns None if it fails to detect the mimetype.

        :param dict value: value to detect its mimetype.

        :rtype: str
        """

        return MIMETypeEnum.JSON

    @property
    def accepted_type(self):
        """
        gets the accepted type for this mimetype handler.

        which could detect mimetype from this type.

        :rtype: type[dict]
        """

        return dict


@mimetype_handler()
class JSONStringMIMETypeHandler(MIMETypeHandlerBase):
    """
    json string mimetype handler class.
    """

    # matches the json inside string.
    # matching is case-insensitive.
    JSON_REGEX = re.compile(r'^[ ]*\{.*\}[ ]*', re.IGNORECASE | re.MULTILINE | re.DOTALL)

    def _mimetype(self, value, **options):
        """
        gets the mimetype of given value.

        returns None if it fails to detect the mimetype.

        :param str value: value to detect its mimetype.

        :rtype: str
        """

        if self.JSON_REGEX.match(value.strip()):
            return MIMETypeEnum.JSON

        return None

    @property
    def accepted_type(self):
        """
        gets the accepted type for this mimetype handler.

        which could detect mimetype from this type.

        :rtype: type[str]
        """

        return str



# -*- coding: utf-8 -*-
"""
encryption services module.
"""

from pyrin.application.services import get_component
from pyrin.security.encryption import EncryptionPackage


def register_encryption_handler(instance, **options):
    """
    registers a new encryption handler or replaces the existing one
    if `replace=True` is provided. otherwise, it raises an error
    on adding an instance which it's name is already available
    in registered handlers.

    :param AbstractEncrypterBase instance: encryption handler to be registered.
                                           it must be an instance of AbstractEncrypterBase.

    :keyword bool replace: specifies that if there is another registered
                           handler with the same name, replace it with
                           the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidEncryptionHandlerTypeError: invalid encryption handler type error.
    :raises InvalidEncryptionHandlerNameError: invalid encryption handler name error.
    :raises DuplicatedEncryptionHandlerError: duplicated encryption handler error.
    """

    return get_component(EncryptionPackage.COMPONENT_NAME).register_encryption_handler(instance,
                                                                                       **options)


def encrypt(text, **options):
    """
    encrypts the given value using specified handler and returns the encrypted result.

    :param str text: text to be encrypted.

    :keyword str handler_name: handler name to be used for encryption.
                               if not provided, default handler from
                               relevant configs will be used.

    :raises EncryptionHandlerNotFoundError: encryption handler not found error.

    :rtype: str
    """

    return get_component(EncryptionPackage.COMPONENT_NAME).encrypt(text, **options)


def decrypt(full_encrypted_value, **options):
    """
    decrypts the given full encrypted value using specified
    handler and returns the decrypted result.

    :param str full_encrypted_value: full encrypted value to be decrypted.

    :raises InvalidEncryptionValueError: invalid encryption value error.
    :raises EncryptionHandlerNotFoundError: encryption handler not found error.
    :raises DecryptionError: decryption error.

    :rtype: str
    """

    return get_component(EncryptionPackage.COMPONENT_NAME).decrypt(full_encrypted_value,
                                                                   **options)


def generate_key(handler_name, **options):
    """
    generates a valid key for the given handler and returns it.

    :param str handler_name: handler name to be used for key generation.

    :keyword int length: the length of generated key in bytes.
                         note that some encryption handlers may not accept custom
                         key length so this value would be ignored on those handlers.

    :raises EncryptionHandlerNotFoundError: encryption handler not found error.

    :rtype: str | tuple[str, str]
    """

    return get_component(EncryptionPackage.COMPONENT_NAME).generate_key(handler_name, **options)



# -*- coding: utf-8 -*-
"""
router structs module.
"""

from werkzeug.routing import Map

from pyrin.core.structs import DTO


class CoreURLMap(Map):
    """
    core url map class.

    this extends the `Map` class to add some functionalities to it.
    """

    def __init__(self, rules=None, default_subdomain="", charset="utf-8",
                 strict_slashes=True, merge_slashes=True, redirect_defaults=True,
                 converters=None, sort_parameters=False, sort_key=None,
                 encoding_errors="replace", host_matching=False):
        """
        initializes an instance of CoreURLMap.

        :param RouteBase rules: sequence of url rules for this map.

        :param str default_subdomain: the default subdomain for rules without
                                      a subdomain defined.

        :param str charset: charset of the url.
                            defaults to `utf-8` if not provided.

        :param bool strict_slashes: if a rule ends with a slash but the matched
                                    url does not, redirect to the url with a trailing slash.

        :param bool merge_slashes: merge consecutive slashes when matching or building
                                   urls. matches will redirect to the normalized url.
                                   slashes in variable parts are not merged.

        :param bool redirect_defaults: this will redirect to the default rule if it
                                       wasn't visited that way. this helps creating
                                       unique urls.

        :param dict converters: a dict of converters that adds additional converters
                                to the list of converters. if you redefine one
                                converter this will override the original one.

        :param bool sort_parameters: if set to `True` the url parameters are sorted.
                                     see `url_encode` for more details.

        :param function sort_key: the sort key function for `url_encode`.
        :param str encoding_errors: the error method to use for decoding.

        :param bool host_matching: if set to `True` it enables the host matching
                                   feature and disables the subdomain one.  if
                                   enabled the `host` parameter to rules is used
                                   instead of the `subdomain` one.
        """

        super().__init__(rules, default_subdomain, charset, strict_slashes,
                         merge_slashes, redirect_defaults, converters,
                         sort_parameters, sort_key, encoding_errors,
                         host_matching)

        # a dict containing the mapping between each url and its available routes.
        # in the form of: {str url: [RouteBase route]}
        self._routes_by_url = DTO()

    def add(self, rule_factory):
        """
        adds a new rule or factory to the map and binds it.

        requires that the route is not bound to another map.

        :param RouteBase | RuleFactory rule_factory: a route or rule factory instance.
        """

        super().add(rule_factory)

        for route in rule_factory.get_rules(self):
            self._routes_by_url.setdefault(route.rule, []).append(route)

    def remove(self, rule_factory):
        """
        removes a rule or factory from the map and unbinds it.

        requires that the rule is bounded to current map.

        :param RouteBase | RuleFactory rule_factory: a `RouteBase` or `RuleFactory`
                                                     instance to be removed from this map.

        :raises RouteIsNotBoundedError: route is not bounded error.
        :raises RouteIsNotBoundedToMapError: route is not bounded to map error.
        """

        for rule in rule_factory.get_rules(self):
            rule.unbind(self)
            self._rules.remove(rule)
            self._remove_by_endpoint(rule)
            self._remove_by_url(rule)

        self._remap = True

    def _remove_by_endpoint(self, route):
        """
        removes given route from endpoint map.

        :param RouteBase route: route instance to be removed from endpoint map.
        """

        routes = self._rules_by_endpoint.get(route.endpoint)
        if routes is None:
            routes = []
        if route in routes:
            routes.remove(route)
            self._rules_by_endpoint[route.endpoint] = routes
        if len(routes) <= 0:
            self._rules_by_endpoint.pop(route.endpoint, None)

    def _remove_by_url(self, route):
        """
        removes given route from url map.

        :param RouteBase route: route instance to be removed from url map.
        """

        routes = self._routes_by_url.get(route.rule)
        if routes is None:
            routes = []
        if route in routes:
            routes.remove(route)
            self._routes_by_url[route.rule] = routes
        if len(routes) <= 0:
            self._routes_by_url.pop(route.rule, None)

    def get_routes_by_endpoint(self, endpoint):
        """
        gets the available routes for given endpoint.

        :param str endpoint: endpoint to get its routes.

        :returns: list[RouteBase]
        :rtype: list
        """

        routes = self._rules_by_endpoint.get(endpoint)
        if routes is None:
            return []

        return routes

    def get_routes_by_url(self, url):
        """
        gets the available routes for given url.

        :param str url: url to get its routes.

        :returns: list[RouteBase]
        :rtype: list
        """

        routes = self._routes_by_url.get(url)
        if routes is None:
            return []

        return routes

    def count_routes_by_endpoint(self, endpoint):
        """
        counts routes for given endpoint.

        :param str endpoint: endpoint to count its routes.

        :rtype: int
        """

        return len(self.get_routes_by_endpoint(endpoint))

    def count_routes_by_url(self, url):
        """
        counts routes for given url.

        :param str url: url to count its routes.

        :rtype: int
        """

        return len(self.get_routes_by_url(url))



# -*- coding: utf-8 -*-
"""
accounting api module.
"""

import pyrin.utils.unique_id as uuid_utils
from demo.accounting.schema import AccountSchema

from pyrin.api.router.decorators import api
from pyrin.core.enumerations import HTTPMethodEnum
from pyrin.core.globals import SECURE_TRUE

import demo.accounting.services as accounting_services

from demo.accounting.models import AccountEntity


# Usage:
# you could implement different api functions here and call corresponding service method this way:
# return accounting_services.method_name(*arg, **kwargs)
from pyrin.database.services import get_current_store


@api('/accounts', methods=HTTPMethodEnum.POST, authenticated=False)
def create(owner=None):
    # uuid = uuid_utils.generate_uuid4()
    # entity = AccountEntity(id=uuid, owner=owner, populate_all=SECURE_TRUE)
    # return entity.save()
    return 1


@api('/accounts', methods=HTTPMethodEnum.GET, authenticated=False, result_schema=AccountSchema())
def get_all():
    store = get_current_store()
    return store.query(AccountEntity.id, AccountEntity.owner).all()


@api('/accounts/<uuid:uuid>', methods=HTTPMethodEnum.GET, authenticated=False)
def get(uuid):
    store = get_current_store()
    return store.query(AccountEntity).filter(AccountEntity.id == uuid).all()



# -*- coding: utf-8 -*-
"""
serializer handlers exceptions module.
"""

from pyrin.core.exceptions import CoreException


class SerializerHandlersException(CoreException):
    """
    serializer handlers exception.
    """
    pass



# -*- coding: utf-8 -*-
"""
json decoder module.
"""

from json.decoder import scanstring
from json.scanner import py_make_scanner

from flask.json import JSONDecoder

import pyrin.globalization.datetime.services as datetime_services

from pyrin.utils.datetime import DEFAULT_DATE_TIME_ISO_REGEX, \
    DEFAULT_DATE_ISO_REGEX, DEFAULT_TIME_ISO_REGEX, DEFAULT_LOCAL_NAIVE_TIME_REGEX, \
    DEFAULT_UTC_ZULU_DATE_TIME_REGEX, DEFAULT_LOCAL_NAIVE_DATE_TIME_REGEX


def scanstring_extended(s, end, strict=True):
    """
    extended scan string method to be able to parse datetime strings.
    """

    s, end = scanstring(s, end, strict)
    if DEFAULT_DATE_TIME_ISO_REGEX.match(s):
        return datetime_services.to_datetime(s, server=False), end
    elif DEFAULT_DATE_ISO_REGEX.match(s):
        return datetime_services.to_date(s), end
    elif DEFAULT_TIME_ISO_REGEX.match(s) or \
            DEFAULT_LOCAL_NAIVE_TIME_REGEX.match(s):
        return datetime_services.to_time(s), end
    elif DEFAULT_UTC_ZULU_DATE_TIME_REGEX.match(s):
        return datetime_services.to_datetime(s, server=False), end
    elif DEFAULT_LOCAL_NAIVE_DATE_TIME_REGEX.match(s):
        return datetime_services.to_datetime(s, server=False, replace_server=False), end
    else:
        return s, end


class CoreJSONDecoder(JSONDecoder):
    """
    the default pyrin json decoder.

    it extends the default flask json decoder to be able to
    convert complex strings to their equivalent python object.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of CoreJSONDecoder.
        """

        super().__init__(*args, **kwargs)
        self.parse_string = scanstring_extended
        self.scan_once = py_make_scanner(self)



# -*- coding: utf-8 -*-
"""
utils datetime module.
"""

import re

from datetime import datetime, timedelta

import pytz
import aniso8601

import pyrin.security.session.services as session_services


# default iso datetime regular expression pattern with utc offset.
# example: '2015-12-24T23:40:15+03:30' or '2015-12-24T23:40:15.8965+03:30'
# length is between 25 to 32.
DEFAULT_DATE_TIME_ISO_REGEX = \
    re.compile(
        r'^[1-2]\d{3}-[0-1]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d(\.\d{1,6})?[+-][0-1]\d:[0-5]\d$')

# default iso date regular expression pattern.
# example: '2015-12-24'
# length is 10.
DEFAULT_DATE_ISO_REGEX = re.compile(r'^[1-2]\d{3}-[0-1]\d-[0-3]\d$')

# default utc datetime regular expression pattern with zulu sign.
# example: '2015-12-24T23:40:15.926Z' or '2015-12-24T23:40:15Z'
# length is between 20 to 27.
DEFAULT_UTC_ZULU_DATE_TIME_REGEX = \
    re.compile(r'^[1-2]\d{3}-[0-1]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d(\.\d{1,6})?Z$')

# default iso time regular expression pattern with utc offset.
# example: '23:40:15+03:30' or '23:40:15.889+03:30'
# length is between 14 to 21.
DEFAULT_TIME_ISO_REGEX = re.compile(r'^[0-2]\d:[0-5]\d:[0-5]\d(\.\d{1,6})?[+-][0-1]\d:[0-5]\d$')

# default local naive datetime regular expression pattern without utc offset.
# example: '2015-12-24T23:40:15' or '2015-12-24T23:40:15.98'
# length is between 19 to 26.
DEFAULT_LOCAL_NAIVE_DATE_TIME_REGEX = \
    re.compile(r'^[1-2]\d{3}-[0-1]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d(\.\d{1,6})?$')

# default local naive time regular expression pattern without utc offset.
# example: '23:40:15' or '23:40:15.565'
# length is between 8 to 15.
DEFAULT_LOCAL_NAIVE_TIME_REGEX = re.compile(r'^[0-2]\d:[0-5]\d:[0-5]\d(\.\d{1,6})?$')


def to_datetime_string(value):
    """
    gets the datetime string representation of input value with utc offset.

    for example: `2015-12-24T23:40:15+03:30`

    :param datetime value: input object to be converted.

    :rtype: str
    """

    return value.isoformat(timespec='seconds')


def to_datetime_string_utc(value):
    """
    gets the datetime string representation of input value in utc with zero offset.

    for example: `2015-12-24T22:40:15+00:00`

    :param datetime value: input object to be converted.

    :rtype: str
    """

    utc_value = normalized_utc(value)
    return to_datetime_string(utc_value)


def to_date_string(value):
    """
    gets the date string representation of input with default format.

    for example: `2015-12-24`

    :param datetime | date value: input object to be converted.

    :rtype: str
    """

    date = value
    if isinstance(value, datetime):
        date = value.date()

    return date.isoformat()


def to_time_string(value):
    """
    gets the time string representation of input datetime with utc offset.

    for example: `23:40:15`

    :param datetime | time value: input object to be converted.

    :rtype: str
    """

    time = value
    if isinstance(value, datetime):
        time = value.timetz()

    return time.isoformat(timespec='seconds')


def to_time_string_utc(value):
    """
    gets the time string representation of input datetime in utc with zero offset.

    for example: `22:40:15`

    :param datetime | time value: input object to be converted.

    :rtype: str
    """

    utc_value = normalized_utc(value)
    return to_time_string(utc_value)


def to_datetime(value):
    """
    converts the input value to it's equivalent python datetime.

    :param str value: string representation of datetime to be converted.

    :rtype: datetime
    """

    return aniso8601.parse_datetime(value)


def to_datetime_utc(value):
    """
    converts the input value to it's equivalent python datetime in utc with zero offset.

    :param str value: string representation of datetime to be converted.

    :rtype: datetime
    """

    converted_value = to_datetime(value)
    return normalized_utc(converted_value)


def to_date(value):
    """
    converts the input value to it's equivalent python date.

    :param str value: string representation of date to be converted.

    :rtype: date
    """

    return aniso8601.parse_date(value)


def to_time(value):
    """
    converts the input value to it's equivalent python time.

    :param str value: string representation of time to be converted.

    :rtype: time
    """

    return aniso8601.parse_time(value)


def utc_now():
    """
    gets current datetime in utc timezone with zero offset.

    :rtype: datetime
    """

    dt_now = datetime.utcnow()
    utc = pytz.utc
    return utc.localize(dt_now)


def normalized_utc(value):
    """
    normalizes input datetime value as utc datetime with zero offset and returns it.

    :param datetime value: value to get it's utc equivalent.

    :rtype: datetime
    """

    utc = pytz.utc
    return utc.normalize(value)


def trunc(value):
    """
    truncates the input datetime value to just date, with default time info.

    it is actually a datetime with time info set to 00:00:00.

    :param datetime value: datetime to be truncated.

    :rtype: datetime
    """

    tz = None
    if isinstance(value, datetime) and value.tzinfo is not None:
        tz = value.tzinfo

    if tz is None:
        current_request = session_services.get_safe_current_request()
        if current_request is not None:
            tz = current_request.timezone

    result = datetime(value.year, value.month, value.day, tzinfo=None)
    if tz is None:
        return result

    return tz.localize(result)


def begin_of_day(value):
    """
    gets a datetime representing the begin of day for given datetime.

    it is actually a datetime with time info set to 00:00:00.

    :param datetime value: value to get its begin of day.

    :rtype: datetime
    """

    return trunc(value)


def end_of_day(value):
    """
    gets a datetime representing the end of day for given datetime.

    it is actually a datetime with time info set to 23:59:59.999999

    :param datetime value: value to get its end of day.

    :rtype: datetime
    """

    return add_days(trunc(value), 1) - timedelta(milliseconds=1)


def add_days(value, days):
    """
    adds the given number of days to specified value and returns it.

    :param datetime value: value to add days to it.

    :param int days: number of days to add.
                     it could be negative if needed.

    :rtype: datetime
    """

    return value + timedelta(days)


def normalize_datetime_range(value_lower, value_upper, **options):
    """
    normalizes the given datetime values range.

    it returns a tuple of two datetime values
    normalized according to given options.

    :param datetime value_lower: lower bound of datetime range.
    :param datetime value_upper: upper bound of datetime range.

    :keyword bool consider_begin_of_day: specifies that consider begin
                                         of day for lower datetime.
                                         defaults to True if not provided.

    :keyword bool consider_end_of_day: specifies that consider end
                                       of day for upper datetime.
                                       defaults to True if not provided.

    :returns: tuple[datetime value_lower: datetime value_upper]
    :rtype: tuple[datetime, datetime]
    """

    consider_begin_of_day = options.get('consider_begin_of_day', True)
    consider_end_of_day = options.get('consider_end_of_day', True)

    # swapping values in case of user mistake.
    if value_upper is not None and value_lower is not None and value_lower > value_upper:
        value_lower, value_upper = value_upper, value_lower

    if value_lower is not None and consider_begin_of_day is True:
        value_lower = begin_of_day(value_lower)

    if value_upper is not None and consider_end_of_day is True:
        value_upper = end_of_day(value_upper)

    return value_lower, value_upper



# -*- coding: utf-8 -*-
"""
province api module.
"""

from pyrin.api.router.decorators import api
from pyrin.core.enumerations import HTTPMethodEnum

import demo.common.province.services as province_services

#
# @api('/provinces/<int:province_id>', methods=HTTPMethodEnum.GET, authenticated=False)
# def get(province_id, **options):
#     """
#     gets the specified province.
#
#     :param int province_id: province id to get its info.
#
#     :raises ProvinceNotFoundError: province not found error.
#
#     :returns: dict(int id,
#                    str name)
#
#     :rtype: dict
#     """
#
#     return province_services.get(province_id, **options)
#
#
# @api('/provinces/find', methods=HTTPMethodEnum.GET, authenticated=False)
# def find(**filters):
#     """
#     finds provinces with given filters.
#
#     :keyword str name: province name.
#
#     :returns: list[dict(int id,
#                         str name)]
#
#     :rtype: list
#     """
#
#     return province_services.find(**filters)
#
#
# @api('/provinces', methods=HTTPMethodEnum.GET, authenticated=False)
# def get_all(**options):
#     """
#     gets all provinces.
#
#     :returns: list[dict(int id,
#                         str name)]
#
#     :rtype: list
#     """
#
#     return province_services.get_all(**options)
#
#
# @api('/provinces', methods=HTTPMethodEnum.POST, authenticated=False)
# def create(name, **options):
#     """
#     creates a province.
#
#     :param str name: province name.
#     """
#
#     return province_services.create(name, **options)



# -*- coding: utf-8 -*-
"""
babel services module.
"""

from pyrin.application.services import get_component
from pyrin.globalization.locale.babel import BabelPackage


def register_cli_handler(instance, **options):
    """
    registers a new babel cli handler or replaces the existing one
    if `replace=True` is provided. otherwise, it raises an error
    on adding a cli handler which is already registered.

    :param BabelCLIHandlerBase instance: babel cli handler to be registered.
                                         it must be an instance of
                                         BabelCLIHandlerBase.

    :keyword bool replace: specifies that if there is another registered
                           cli handler with the same name, replace it
                           with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidCLIHandlerTypeError: invalid cli handler type error.
    :raises DuplicatedCLIHandlerError: duplicated cli handler error.
    """

    get_component(BabelPackage.COMPONENT_NAME).register_cli_handler(instance, **options)


def execute(handler_name, **options):
    """
    executes the handler with the given name with given inputs.

    :param str handler_name: handler name to be executed.

    :raises CLIHandlerNotFoundError: cli handler not found error.
    """

    return get_component(BabelPackage.COMPONENT_NAME).execute(handler_name, **options)


def enable(include_pyrin=True, include_app=True):
    """
    enables locale management for the application.

    :param bool include_pyrin: specifies that it should extract pyrin localizable
                               messages. defaults to True if not provided.

    :param bool include_app: specifies that it should extract application
                             localizable messages. defaults to True if not provided.
    """

    return get_component(BabelPackage.COMPONENT_NAME).enable(include_pyrin, include_app)


def rebuild(include_pyrin=True, include_app=True, locale=None):
    """
    it will do the three complete steps needed to
    update and compile locales with new messages.

    it will do:
        1. extract
        2. update
        3. compile

    this command is defined for convenient of usage, but if you need
    to do these steps separately, you could ignore this command and
    use the relevant command for each step.

    :keyword bool include_pyrin: specifies that it must extract pyrin localizable
                                 messages as well. defaults to True if not provided.

    :keyword bool include_app: specifies that it must extract application localizable
                               messages as well. defaults to True if not provided.

    :keyword str locale: locale name of the catalog to compile.
                         it will compile all catalogs if not provided.
    """

    return get_component(BabelPackage.COMPONENT_NAME).rebuild(include_pyrin,
                                                              include_app, locale)


def check_init(locale):
    """
    checks that locale with given name does not exist.

    :param str locale: locale name for the new localized catalog.
                       for example: `en` or `fr` or ...

    :raises LocaleAlreadyExistedError: locale already existed error.
    """

    return get_component(BabelPackage.COMPONENT_NAME).check_init(locale)


def get_package_class():
    """
    gets the package class of babel manager.

    :raises PackageClassIsNotSetError: package class is not set error.

    :returns: type[BabelPackage]
    """

    return get_component(BabelPackage.COMPONENT_NAME).get_package_class()



# -*- coding: utf-8 -*-
"""
database migration package.
"""

from pyrin.packaging.base import Package


class DatabaseMigrationPackage(Package):
    """
    database migration package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'database.migration.component'



# -*- coding: utf-8 -*-
"""
permission base module.
"""

from pyrin.core.globals import SECURE_TRUE
from pyrin.security.permission.base import PermissionBase

from demo.security.permission.models import PermissionEntity


class CorePermission(PermissionBase):
    """
    core permission class.
    all application permissions must be an instance of this.
    """

    def __init__(self, permission_id, description, **options):
        """
        initializes an instance of CorePermission.

        :param int permission_id: permission id.
        :param str description: permission description.
        """

        self.id = permission_id
        self.description = description

        super().__init__(**options)

    def __hash__(self):
        return hash(self.get_id())

    def __eq__(self, other):
        if not isinstance(other, CorePermission):
            return False

        return other.get_id() == self.get_id()

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return '{id}-{description}'.format(id=self.get_id(), description=self.description)

    def __repr__(self):
        return str(self)

    def to_entity(self):
        """
        gets the equivalent entity of current permission.

        :rtype: PermissionEntity
        """

        entity = PermissionEntity(id=self.id, description=self.description,
                                  populate_all=SECURE_TRUE)
        return entity

    def get_id(self):
        """
        gets permission id.
        note that this object must be fully unique for each different permission.

        :rtype: str
        """

        return self.id



# -*- coding: utf-8 -*-
"""
logging masking package.
"""

from pyrin.packaging.base import Package


class LoggingMaskingPackage(Package):
    """
    logging masking package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.configuration']
    COMPONENT_NAME = 'logging.masking.component'
    CONFIG_STORE_NAMES = ['logging.masking']



# -*- coding: utf-8 -*-
"""
mimetype decorators module.
"""

import pyrin.processor.mimetype.services as mimetype_services


def mimetype_handler(*args, **kwargs):
    """
    decorator to register a mimetype handler.

    :param object args: mimetype handler class constructor arguments.
    :param object kwargs: mimetype handler class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           mimetype handler with the same name and accepted type,
                           replace it with the new one, otherwise raise
                           an error. defaults to False.

    :raises InvalidMIMETypeHandlerTypeError: invalid mimetype handler type error.
    :raises DuplicatedMIMETypeHandlerError: duplicated mimetype handler error.

    :returns: mimetype handler class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available mimetype handlers.

        :param type cls: mimetype handler class.

        :returns: mimetype handler class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        mimetype_services.register_mimetype_handler(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
cli core command module.
"""

import sys

import pyrin.cli.core.services as core_services

from pyrin.utils.custom_print import print_error


def main():
    try:
        handler_name = sys.argv[1]
        core_services.create(handler_name)
    except IndexError:
        print_error('Please input the command name.', force=True)
    except Exception as error:
        print_error(error, force=True)



# -*- coding: utf-8 -*-
"""
city package.
"""

from pyrin.packaging.base import Package


class CityPackage(Package):
    """
    city package class.
    """

    NAME = __name__
    DEPENDS = ['demo.common.province']
    COMPONENT_NAME = 'common.city.component'



# -*- coding: utf-8 -*-
"""
string normalizer decorators module.
"""

import pyrin.utilities.string.normalizer.services as normalizer_services


def string_normalizer(*args, **kwargs):
    """
    decorator to register a string normalizer.

    :param object args: normalizer class constructor arguments.
    :param object kwargs: normalizer class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           normalizer with the same name, replace it with
                           the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidStringNormalizerTypeError: invalid string normalizer type error.
    :raises DuplicatedStringNormalizerError: duplicated string normalizer error.

    :returns: string normalizer class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available string normalizers.

        :param type cls: normalizer class.

        :returns: normalizer class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        normalizer_services.register_normalizer(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
accounting component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component

from demo.accounting import AccountingPackage
from demo.accounting.manager import AccountingManager


@component(AccountingPackage.COMPONENT_NAME)
class AccountingComponent(Component, AccountingManager):
    """
    accounting component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
caching audit component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.caching.audit import CachingAuditPackage
from pyrin.caching.audit.manager import CachingAuditManager


@component(CachingAuditPackage.COMPONENT_NAME)
class CachingAuditComponent(Component, CachingAuditManager):
    """
    caching audit component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
authentication manager module.
"""

import pyrin.security.session.services as session_services

from pyrin.security.authentication.exceptions import AuthenticationFailedError
from pyrin.security.authentication.manager import AuthenticationManager \
    as BaseAuthenticationManager


class AuthenticationManager(BaseAuthenticationManager):
    """
    authentication manager class.
    """

    def _push_custom_data(self, header, payload, **options):
        """
        pushes the custom data into current request from input values.

        :param dict header: token header data.

        :param dict payload: payload data of authenticated token.
        :type payload: dict(str type: token type,
                            int user_id: user id)

        :raises InvalidUserError: invalid user error.
        :raises CouldNotOverwriteCurrentUserError: could not overwrite current user error.
        """

        user_id = payload.pop('user_id', None)
        self._push_component_custom_key(user_id)
        session_services.set_current_user(user_id)

    def _validate_custom(self, header, payload, **options):
        """
        validates the given inputs for custom attributes.
        an error will be raised if validation fails.

        :param dict header: token header data.

        :param dict payload: payload data to be validated.
        :type payload: dict(str type: token type,
                            int user_id: user id)

        :raises AuthenticationFailedError: authentication failed error.
        """

        user_id = payload.get('user_id', 0)
        if user_id <= 0:
            raise AuthenticationFailedError('User id must be provided in payload.')



# -*- coding: utf-8 -*-
"""
province package.
"""

from pyrin.packaging.base import Package


class ProvincePackage(Package):
    """
    province package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'common.province.component'



# -*- coding: utf-8 -*-
"""
serializer handlers list module.
"""

import pyrin.converters.serializer.services as serializer_services
import pyrin.configuration.services as config_services

from pyrin.converters.serializer.decorators import serializer
from pyrin.converters.serializer.handlers.base import SerializerBase


@serializer()
class ListSerializer(SerializerBase):
    """
    list serializer class.
    """

    def __init__(self, *args, **options):
        """
        initializes an instance of ListSerializer.

        :param object args: constructor arguments.
        """

        super().__init__(*args, **options)

        self._default_index_name = config_services.get('api', 'schema', 'index_name')
        self._default_start_index = config_services.get('api', 'schema', 'start_index')

    def _serialize(self, value, **options):
        """
        serializes the given value.

        returns serialized list on success or `NULL` object if serialization fails.

        :param list[object] value: list value to be serialized.

        :keyword bool indexed: specifies that list results must
                               include an extra field as row index.
                               the name of the index field and the initial value
                               of index could be provided by `index_name` and
                               `start_index` respectively.

        :keyword str index_name: name of the extra field to contain
                                 the row index of each result.

        :keyword int start_index: the initial value of row index.

        :keyword ResultSchema result_schema: result schema instance to be
                                             used to create computed columns.
                                             defaults to None if not provided.

        :returns: serialized list of objects
        :rtype: list[dict]
        """

        if len(value) <= 0:
            return []

        indexed = options.get('indexed', False)
        index_name = options.get('index_name', self._default_index_name)
        start_index = options.get('start_index', self._default_start_index)
        result = []

        # if indexing is requested, all items must be a dict after serialization.
        # so we just check the first item to be a dict after serialization, otherwise
        # no indexing will be done.
        first_item = serializer_services.serialize(value[0], **options)
        if indexed is False or not isinstance(first_item, dict):
            indexed = False

        if indexed is True:
            first_item[index_name] = start_index
            start_index = start_index + 1

        result.append(first_item)
        value = value[1:]

        # we set 'indexed=False' to prevent inner lists from indexing.
        options.update(indexed=False)
        for item in value:
            converted = serializer_services.serialize(item, **options)
            if indexed is True:
                converted[index_name] = start_index
                start_index = start_index + 1

            result.append(converted)

        return result

    @property
    def accepted_type(self):
        """
        gets the accepted type for this serializer.

        which could serialize values from this type.

        :rtype: type[list]
        """

        return list



# -*- coding: utf-8 -*-
"""
babel exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class BabelHandlersException(CoreException):
    """
    babel handlers exception.
    """
    pass


class BabelHandlersBusinessException(CoreBusinessException, BabelHandlersException):
    """
    babel handlers business exception.
    """
    pass


class LocaleAlreadyExistedError(BabelHandlersBusinessException):
    """
    locale already existed error.
    """
    pass



# -*- coding: utf-8 -*-
"""
users component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component

from demo.security.users import UsersPackage
from demo.security.users.manager import UsersManager


@component(UsersPackage.COMPONENT_NAME, replace=True)
class UsersComponent(Component, UsersManager):
    """
    users component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
encryption package.
"""

from pyrin.packaging.base import Package


class EncryptionPackage(Package):
    """
    encryption package class.
    """

    NAME = __name__
    DEPENDS = []
    COMPONENT_NAME = 'security.encryption.component'



# -*- coding: utf-8 -*-
"""
security component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component

from demo.security import SecurityPackage
from demo.security.manager import SecurityManager


@component(SecurityPackage.COMPONENT_NAME, replace=True)
class SecurityComponent(Component, SecurityManager):
    """
    security component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
hashing manager module.
"""

import pyrin.configuration.services as config_services

from pyrin.core.structs import Context, Manager
from pyrin.security.hashing import HashingPackage
from pyrin.security.hashing.interface import AbstractHashingBase
from pyrin.utils.custom_print import print_warning
from pyrin.security.hashing.exceptions import InvalidHashingHandlerTypeError, \
    InvalidHashingHandlerNameError, DuplicatedHashingHandlerError, InvalidHashError, \
    HashingHandlerNotFoundError


class HashingManager(Manager):
    """
    hashing manager class.
    """

    package_class = HashingPackage

    def __init__(self):
        """
        initializes an instance of HashingManager.
        """

        super().__init__()

        self._hashing_handlers = Context()
        self._separator = '$'

    def register_hashing_handler(self, instance, **options):
        """
        registers a new hashing handler or replaces the existing one
        if `replace=True` is provided. otherwise, it raises an error
        on adding an instance which it's name is already available
        in registered handlers.

        :param AbstractHashingBase instance: hashing handler to be registered.
                                             it must be an instance of AbstractHashingBase.

        :keyword bool replace: specifies that if there is another registered
                               handler with the same name, replace it with
                               the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidHashingHandlerTypeError: invalid hashing handler type error.
        :raises InvalidHashingHandlerNameError: invalid hashing handler name error.
        :raises DuplicatedHashingHandlerError: duplicated hashing handler error.
        """

        if not isinstance(instance, AbstractHashingBase):
            raise InvalidHashingHandlerTypeError('Input parameter [{instance}] is '
                                                 'not an instance of [{base}].'
                                                 .format(instance=instance,
                                                         base=AbstractHashingBase))

        if instance.get_name() is None or len(instance.get_name().strip()) == 0:
            raise InvalidHashingHandlerNameError('Hashing handler [{instance}] '
                                                 'does not have a valid name.'
                                                 .format(instance=instance))

        # checking whether is there any registered instance with the same name.
        if instance.get_name() in self._hashing_handlers:
            replace = options.get('replace', False)

            if replace is not True:
                raise DuplicatedHashingHandlerError('There is another registered hashing '
                                                    'handler with name [{name}] but "replace" '
                                                    'option is not set, so handler '
                                                    '[{instance}] could not be registered.'
                                                    .format(name=instance.get_name(),
                                                            instance=instance))

            old_instance = self._hashing_handlers[instance.get_name()]
            print_warning('Hashing handler [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance, new_instance=instance))

        # registering new hashing handler.
        self._hashing_handlers[instance.get_name()] = instance

    def generate_hash(self, text, **options):
        """
        gets the hash of input text using a random or specified salt.

        :param str text: text to be hashed.

        :keyword str handler_name: handler name to be used for hash generation.
                                   if not provided, default handler from
                                   relevant configs will be used.

        :keyword bytes salt: salt to be used for hashing.
                             if not provided, a random salt will be generated
                             considering `salt_length` option.

        :keyword str internal_algorithm: internal algorithm to be used
                                         for hashing. if not provided,
                                         default value from relevant
                                         config will be used.

        :keyword int rounds: rounds to perform for generating hash.
                             if not provided, default value from
                             relevant config will be used.

        :keyword int salt_length: salt length to be used for hashing.
                                  if `salt` option is provided, then
                                  this value will be ignored.
                                  if not provided, default value from
                                  relevant config will be used.

        :keyword str prefix: prefix to be used for bcrypt hashing.

        :rtype: str
        """

        return self._get_hashing_handler(**options).generate_hash(text, **options)

    def is_match(self, text, full_hashed_value, **options):
        """
        gets a value indicating that given text's
        hash is identical to given full hashed value.

        :param str text: text to be hashed.
        :param str full_hashed_value: full hashed value to compare with.

        :rtype: bool
        """

        handler_name = self._extract_handler_name(full_hashed_value, **options)
        return self._get_hashing_handler(handler_name=handler_name).is_match(
            text, full_hashed_value, **options)

    def _get_hashing_handler(self, **options):
        """
        gets the specified hashing handler.

        :keyword str handler_name: handler name to get.
                                   if not provided, default handler from
                                   relevant configs will be used.

        :raises HashingHandlerNotFoundError: hashing handler not found error.

        :rtype: AbstractHashingBase
        """

        handler_name = options.get('handler_name', self._get_default_handler_name())
        if handler_name not in self._hashing_handlers.keys():
            raise HashingHandlerNotFoundError('Hashing handler [{name}] not found.'
                                              .format(name=handler_name))

        return self._hashing_handlers[handler_name]

    def _get_default_handler_name(self):
        """
        gets default hashing handler name from configs.

        :rtype: str
        """

        return config_services.get('security', 'hashing', 'default_hashing_handler')

    def _extract_handler_name(self, full_hashed_value, **options):
        """
        extracts the handler name of given full hashed value.

        :param str full_hashed_value: full hashed value to extract the handler name from.

        :raises InvalidHashError: invalid hash error.

        :rtype: str
        """

        if full_hashed_value.count(self._separator) < 2:
            raise InvalidHashError('Input hash value has an incorrect format.')

        items = full_hashed_value.split(self._separator, 2)
        return items[1]



# -*- coding: utf-8 -*-
"""
celery package.
"""

from pyrin.packaging.base import Package


class CeleryPackage(Package):
    """
    celery package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'task_queues.celery.component'
    CONFIG_STORE_NAMES = ['celery']
    DEPENDS = ['pyrin.configuration',
               'pyrin.logging']



# -*- coding: utf-8 -*-
"""
province exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class ProvinceManagerException(CoreException):
    """
    province manager exception.
    """
    pass


class ProvinceManagerBusinessException(CoreBusinessException,
                                       ProvinceManagerException):
    """
    province manager business exception.
    """
    pass


class ProvinceNotFoundError(ProvinceManagerBusinessException):
    """
    province not found error.
    """
    pass



# -*- coding: utf-8 -*-
"""
router decorators module.
"""

import pyrin.api.router.services as router_services


def api(url, methods=None, authenticated=True, permissions=None, **options):
    """
    decorator to register an api handler for application.

    this decorator could take all the options that are used in route initialization.

    :param str url: the url rule as string.

    :param str | tuple[str] methods: http methods that this rule should handle.
                                     if not provided, defaults to `GET`.

    :param bool authenticated: specifies that this route could not be accessed
                               if the requester has not been authenticated.
                               defaults to True if not provided.

    :param PermissionBase | tuple[PermissionBase] permissions: all required permissions
                                                               for accessing this route.

    :keyword bool fresh_auth: specifies that this route could not be accessed
                              if the requester has not a fresh authentication.
                              fresh authentication means an authentication that
                              has been done by providing user credentials to
                              server. defaults to False if not provided.

    :keyword bool replace: specifies that this route must replace any existing
                           route with the same url and http methods or raise
                           an error if not provided. defaults to False.

    :keyword int max_content_length: max content length that this route could handle,
                                     in bytes. if not provided, it will be set to
                                     `restricted_max_content_length` api config key.
                                     note that this value should be lesser than or equal
                                     to `max_content_length` api config key, otherwise
                                     it will cause an error.

    :keyword int status_code: status code to be returned on successful responses.
                              defaults to corresponding status code of request's
                              http method if not provided.

    :note status_code: it could be a value from `InformationResponseCodeEnum`
                       or `SuccessfulResponseCodeEnum` or `RedirectionResponseCodeEnum`.

    :keyword bool strict_status: specifies that it should only consider
                                 the status code as processed if it is from
                                 `InformationResponseCodeEnum` or
                                 `SuccessfulResponseCodeEnum` or
                                 `RedirectionResponseCodeEnum` values. otherwise
                                 all codes from `INFORMATION_CODE_MIN`
                                 to `INFORMATION_CODE_MAX` or from
                                 `SUCCESS_CODE_MIN` to `SUCCESS_CODE_MAX`
                                 or from `REDIRECTION_CODE_MIN` to
                                 `REDIRECTION_CODE_MAX` will be considered
                                 as processed. defaults to True if not provided.

    :keyword str | list[str] environments: a list of all environments that this
                                           route must be exposed on them.
                                           the values could be from all available
                                           environments in environments config store.
                                           for example: `production`, `development`.
                                           if not provided, the route will be exposed
                                           on all environments.

    :keyword ResultSchema result_schema: result schema to be used to filter results.

    :keyword bool indexed: specifies that list results must
                           include an extra field as row index.
                           the name of the index field and the initial value
                           of index could be provided by `index_name` and
                           `start_index` respectively. `indexed` keyword has
                           only effect if the returning result contains a list
                           of objects.

    :keyword str index_name: name of the extra field to contain
                             the row index of each result. if not provided
                             defaults to `row_num` value.

    :keyword int start_index: the initial value of row index. if not
                              provided, starts from 1.

    :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                  which has `allow_read=False` or its name
                                                  starts with underscore `_`, should not
                                                  be included in result dict. defaults to
                                                  `SECURE_TRUE` if not provided. it will
                                                  be used only for entity conversion. this
                                                  value will override the corresponding
                                                  value of `result_schema` if provided.

    :keyword int depth: a value indicating the depth for conversion.
                        for example if entity A has a relationship with
                        entity B and there is a list of B in A, if `depth=0`
                        is provided, then just columns of A will be available
                        in result dict, but if `depth=1` is provided, then all
                        B entities in A will also be included in the result dict.
                        actually, `depth` specifies that relationships in an
                        entity should be followed by how much depth.
                        note that, if `columns` is also provided, it is required to
                        specify relationship property names in provided columns.
                        otherwise they won't be included even if `depth` is provided.
                        defaults to `default_depth` value of database config store.
                        please be careful on increasing `depth`, it could fail
                        application if set to higher values. choose it wisely.
                        normally the maximum acceptable `depth` would be 2 or 3.
                        there is a hard limit for max valid `depth` which is set
                        in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                        `depth` value than this limit, will cause an error.
                        it will be used only for entity conversion.
                        this value will override the corresponding value of
                        `result_schema` if provided.

    :keyword bool no_cache: a value indicating that the response returning from this route
                            must have a `Cache-Control: no-cache` header. this header will
                            be automatically added. defaults to False if not provided.

    :keyword int request_limit: number of allowed requests to this
                                route before it unregisters itself.
                                defaults to None if not Provided.

    :keyword int lifetime: number of seconds in which this route must remain
                           responsive after initial registration. after this
                           period, the route will unregister itself.
                           defaults to None if not provided.

    :note request_limit, lifetime: if both of these values are provided, this
                                   route will unregister itself if any of
                                   these two conditions are met.

    :keyword bool paged: specifies that this route should return paginated results.
                         defaults to False if not provided.

    :keyword int page_size: default page size for this route.
                            defaults to `default_page_size` from
                            `database` config store if not provided.

    :keyword int max_page_size: maximum page size that client is allowed
                                to request for this route. defaults to
                                `max_page_size` from `database` configs store
                                if not provided.

    :keyword bool cors_enabled: specifies that cross origin resource sharing is enabled.
                                if not provided, it will be get from cors config store.

    :keyword bool cors_always_send: specifies that cors headers must be included in
                                    response even if the request does not have origin header.
                                    if not provided, it will be get from cors config store.

    :keyword list[str] cors_allowed_origins: a list of extra allowed origins to be used
                                             in conjunction with default allowed ones.

    :keyword list[str] cors_exposed_headers: extra exposed headers to be combined
                                             with default ones.

    :keyword list[str] cors_allowed_headers: extra allowed headers to be combined
                                             with default ones.

    :keyword bool cors_allow_credentials: specifies that browsers are allowed to pass
                                          response headers to front-end javascript code
                                          if the route is authenticated.
                                          if not provided, it will be get from cors config
                                          store.

    :keyword int cors_max_age: maximum number of seconds to cache results.
                               if not provided, it will be get from cors config store.

    :keyword bool provide_automatic_options: controls whether the `OPTIONS` method should be
                                             added automatically.
                                             this can also be controlled by setting the
                                             `view_func.provide_automatic_options = False`
                                             before adding the rule.

    :keyword dict defaults: an optional dict with defaults for other rules with the
                            same endpoint. this is a bit tricky but useful if you
                            want to have unique urls.

    :keyword str subdomain: the subdomain rule string for this rule. If not specified the
                            rule only matches for the `default_subdomain` of the map. if
                            the map is not bound to a subdomain this feature is disabled.

    :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only for
                                  this rule. if not specified the `Map` setting is used.

    :keyword bool merge_slashes: override `Map.merge_slashes` for this rule.

    :keyword bool build_only: set this to True and the rule will never match but will
                              create a url that can be build. this is useful if you have
                              resources on a subdomain or folder that are not handled by
                              the WSGI application (like static data)

    :keyword str | callable redirect_to: if given this must be either a string
                                         or callable. in case of a callable it's
                                         called with the url adapter that
                                         triggered the match and the values
                                         of the url as keyword arguments and has
                                         to return the target for the redirect,
                                         otherwise it has to be a string with
                                         placeholders in rule syntax.

    :keyword bool alias: if enabled this rule serves as an alias for another rule with
                         the same endpoint and arguments.

    :keyword str host: if provided and the url map has host matching enabled this can be
                       used to provide a match rule for the whole host. this also means
                       that the subdomain feature is disabled.

    :keyword bool websocket: if set to True, this rule is only matches for
                             websocket (`ws://`, `wss://`) requests. by default,
                             rules will only match for http requests.
                             defaults to False if not provided.

    :raises DuplicateRouteURLError: duplicate route url error.
    :raises OverwritingEndpointIsNotAllowedError: overwriting endpoint is not allowed error.
    :raises PageSizeLimitError: page size limit error.
    :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
    :raises InvalidViewFunctionTypeError: invalid view function type error.
    :raises InvalidResultSchemaTypeError: invalid result schema type error.
    :raises InvalidResponseStatusCodeError: invalid response status code error.

    :rtype: function
    """

    def decorator(func):
        """
        decorates the given function and registers it as an api handler.

        :param function func: function to register it as an api handler.

        :rtype: function
        """

        router_services.add_route(url, view_func=func, methods=methods,
                                  authenticated=authenticated,
                                  permissions=permissions, **options)

        return func

    return decorator



# -*- coding: utf-8 -*-
"""
caching audit package.
"""

from pyrin.packaging.base import Package


class CachingAuditPackage(Package):
    """
    caching audit package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'caching.audit.component'
    DEPENDS = ['pyrin.audit']



# -*- coding: utf-8 -*-
"""
city services module.
"""

from pyrin.application.services import get_component

from demo.common.city import CityPackage
from pyrin.caching.remote.decorators import redis


def get(city_id, **options):
    """
    gets the specified city.

    :param int city_id: city id to get its info.

    :raises CityNotFoundError: city not found error.

    :returns: dict(int id,
                   str name,
                   int province_id: province id)

    :rtype: dict
    """

    return get_component(CityPackage.COMPONENT_NAME).get(city_id, **options)


def clear():
    return get_component(CityPackage.COMPONENT_NAME).clear()


@redis(expire=100000)
def find(name, **filters):
    """
    finds cities with given filters.

    :keyword str name: city name.
    :keyword int province_id: province id.

    :returns: list[dict(int id,
                        str name,
                        int province_id: province id)]

    :rtype: list
    """

    return get_component(CityPackage.COMPONENT_NAME).find(name, **filters)


def create(**options):
    """
    creates a city.

    :param str name: city name.
    :param int province_id: province id.
    """

    return get_component(CityPackage.COMPONENT_NAME).create(**options)



# -*- coding: utf-8 -*-
"""
security package.
"""

from pyrin.security import SecurityPackage as BaseSecurityPackage


class SecurityPackage(BaseSecurityPackage):
    """
    security package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
security models module.
"""

from sqlalchemy import Unicode, DateTime, Integer, Boolean

from pyrin.database.model.base import CoreEntity
from pyrin.database.orm.sql.schema.base import CoreColumn


class UserBaseEntity(CoreEntity):
    """
    user base entity class.
    """

    _table = 'user'

    id = CoreColumn(name='id', type_=Integer, autoincrement=True,
                    primary_key=True, index=True)


class UserEntity(UserBaseEntity):
    """
    user entity class.
    """

    _extend_existing = True

    username = CoreColumn(name='username', type_=Unicode(50), nullable=False, unique=True)
    password_hash = CoreColumn(name='password_hash', type_=Unicode(250),
                               nullable=False, allow_read=False, allow_write=False)
    last_login_date = CoreColumn(name='last_login_date', type_=DateTime(timezone=True))
    first_name = CoreColumn(name='first_name', type_=Unicode(50), nullable=False)
    last_name = CoreColumn(name='last_name', type_=Unicode(50), nullable=False)
    is_active = CoreColumn(name='is_active', type_=Boolean, nullable=False, default=True)



# -*- coding: utf-8 -*-
"""
city component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component

from demo.common.city import CityPackage
from demo.common.city.manager import CityManager


@component(CityPackage.COMPONENT_NAME)
class CityComponent(Component, CityManager):
    """
    city component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
province models module.
"""

from sqlalchemy import Unicode, Integer

from pyrin.database.model.base import CoreEntity
from pyrin.database.orm.sql.schema.base import CoreColumn


class ProvinceBaseEntity(CoreEntity):
    """
    province base entity class.
    """

    _table = 'province'

    id = CoreColumn(name='id', autoincrement=True, type_=Integer, primary_key=True)


class ProvinceEntity(ProvinceBaseEntity):
    """
    province entity class.
    """

    _extend_existing = True

    name = CoreColumn(name='name', type_=Unicode, unique=True)



# -*- coding: utf-8 -*-
"""
security utils manager module.
"""

import secrets

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization

import pyrin.configuration.services as config_services

from pyrin.core.structs import Manager
from pyrin.security.utils import SecurityUtilsPackage
from pyrin.settings.static import APPLICATION_ENCODING


class SecurityUtilsManager(Manager):
    """
    security utils manager class.
    """

    package_class = SecurityUtilsPackage

    def generate_rsa_key(self, length=None, **options):
        """
        generates a pair of public/private rsa keys.

        :param int length: key length in bits.
                           if not provided, `rsa_default_key_length`
                           config will be used.

        :returns: tuple[str public_key, str private_key]
        :rtype: tuple[str, str]
        """

        if length is None:
            length = config_services.get('security', 'general', 'rsa_default_key_length')

        key = rsa.generate_private_key(public_exponent=65537,
                                       key_size=length,
                                       backend=default_backend())

        private = key.private_bytes(encoding=serialization.Encoding.PEM,
                                    format=serialization.PrivateFormat.TraditionalOpenSSL,
                                    encryption_algorithm=serialization.NoEncryption())

        public_part = key.public_key()
        public = public_part.public_bytes(encoding=serialization.Encoding.PEM,
                                          format=serialization.PublicFormat.SubjectPublicKeyInfo)

        return public.decode(APPLICATION_ENCODING), private.decode(APPLICATION_ENCODING)

    def load_rsa_key(self, public_pem, private_pem, **options):
        """
        loads public/private rsa key objects from provided value.

        :param str public_pem: public key content to load from.
        :param str private_pem: private key content to load from.

        :returns: tuple[object public_key, object private_key]
        :rtype: tuple[object, object]
        """

        private_key = serialization.load_pem_private_key(private_pem.encode(APPLICATION_ENCODING),
                                                         password=None,
                                                         backend=default_backend())

        public_key = serialization.load_pem_public_key(public_pem.encode(APPLICATION_ENCODING),
                                                       backend=default_backend())

        return public_key, private_key

    def get_bytes(self, length=None, **options):
        """
        gets a secure random bytes with given length.

        the result value should not be decoded to string, because
        it's not a safe-string and may cause an error.
        if you want string representation, use `get_hex` or `get_url_safe` methods.

        :param int length: length of random bytes to be get.
                           if not provided, `default_secure_random_size`
                           config will be used.

        :rtype: bytes
        """

        if length is None:
            length = config_services.get('security', 'general',
                                         'default_secure_random_size')

        return secrets.token_bytes(length)

    def get_hex(self, length=None, **options):
        """
        gets a secure random hex string with given length.

        :param int length: length of random string to be get in bytes.
                           if not provided, `default_secure_random_size`
                           config will be used.

        :rtype: str
        """

        if length is None:
            length = config_services.get('security', 'general',
                                         'default_secure_random_size')

        return secrets.token_hex(length)

    def get_url_safe(self, length=None, **options):
        """
        gets a secure random url-safe string with given length.

        :param int length: length of random string to be get in bytes.
                           if not provided, `default_secure_random_size`
                           config will be used.

        :rtype: str
        """

        if length is None:
            length = config_services.get('security', 'general',
                                         'default_secure_random_size')

        return secrets.token_urlsafe(length)



# -*- coding: utf-8 -*-
"""
serializer handlers row_result module.
"""

import pyrin.api.schema.services as schema_services

from pyrin.core.structs import DTO
from pyrin.core.globals import ROW_RESULT
from pyrin.database.model.base import BaseEntity
from pyrin.converters.serializer.decorators import serializer
from pyrin.converters.serializer.handlers.base import SerializerBase


@serializer()
class RowResultSerializer(SerializerBase):
    """
    row result serializer class.
    """

    def _serialize(self, value, **options):
        """
        serializes the given value.

        :param ROW_RESULT value: row result value to be serialized.

        :keyword list[str] columns: column names to be included in result.
                                    for example:
                                    `columns=['id', 'name', 'age']`
                                    if provided column names are not
                                    available in result, they will be ignored.

        :keyword dict[str, str] rename: column names that must be renamed in the result.
                                        it must be a dict with keys as original column
                                        names and values as new column names that should
                                        be exposed instead of original column names.
                                        for example:
                                        `rename=dict(age='new_age', name='new_name')`
                                        if provided rename columns are not available in
                                        result, they will be ignored.

        :keyword list[str] exclude: column names to be excluded from
                                    result. for example:
                                    `exclude=['id', 'name', 'age']`
                                    if provided excluded columns are not
                                    available in result, they will be
                                    ignored.

        :keyword ResultSchema result_schema: result schema instance to be
                                             used to create computed columns.
                                             defaults to None if not provided.

        :rtype: dict
        """

        if len(value) <= 0:
            return DTO()

        requested_columns, rename, excluded_columns = self._extract_conditions(**options)
        base_columns = value.keys()
        entities = DTO()
        to_remove_keys = []
        for item in base_columns:
            instance = getattr(value, item)
            if isinstance(instance, BaseEntity):
                computed_entity_columns = schema_services.get_computed_entity_columns(instance,
                                                                                      **options)
                serialized_entity = instance.to_dict(**options)
                serialized_entity.update(computed_entity_columns)
                entities[item] = serialized_entity
                to_remove_keys.append(item)
                excluded_columns.append(item)

        computed_row_columns = schema_services.get_computed_row_columns(value, **options)

        if len(requested_columns) <= 0 and \
                len(rename) <= 0 and len(excluded_columns) <= 0:

            result = DTO(zip(base_columns, value))
            result.update(computed_row_columns)
            return result

        for item in to_remove_keys:
            base_columns.remove(item)

        if len(requested_columns) > 0:
            requested_columns = set(requested_columns).intersection(set(base_columns))
        else:
            requested_columns = base_columns

        result = DTO()
        for key, entity in entities.items():
            result.update(entity)

        for col in requested_columns:
            if col not in excluded_columns:
                result[rename.get(col, col)] = getattr(value, col)

        result.update(computed_row_columns)
        return result

    @property
    def accepted_type(self):
        """
        gets the accepted type for this serializer.

        which could serialize values from this type.

        :rtype: type[ROW_RESULT]
        """

        return ROW_RESULT

    def _extract_conditions(self, **options):
        """
        extracts all conditions available in given options.

        it extracts columns, rename and exclude values.

        :keyword list[str] columns: column names to be included in result.
                                    for example:
                                    `columns=['id', 'name', 'age']`
                                    if provided column names are not
                                    available in result, they will be ignored.

        :keyword dict[str, str] rename: column names that must be renamed in the result.
                                        it must be a dict with keys as original column
                                        names and values as new column names that should
                                        be exposed instead of original column names.
                                        for example:
                                        `rename=dict(age='new_age', name='new_name')`
                                        if provided rename columns are not available in
                                        result, they will be ignored.

        :keyword list[str] exclude: column names to be excluded from
                                    result. for example:
                                    `exclude=['id', 'name', 'age']`
                                    if provided excluded columns are not
                                    available in result, they will be
                                    ignored.

        :returns: tuple[list[str column_name],
                        dict[str original_column, str new_column],
                        list[str excluded_column]]

        :rtype: tuple[list[str], dict[str, str], list[str]]
        """

        columns = options.get('columns', None)
        rename = options.get('rename', None)
        exclude = options.get('exclude', None)

        if columns is None:
            columns = []

        if rename is None:
            rename = {}

        if exclude is None:
            exclude = []

        return columns, rename, exclude



# -*- coding: utf-8 -*-
"""
serializer handlers base module.
"""

from abc import abstractmethod

from pyrin.converters.serializer.interface import AbstractSerializerBase
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.globals import NULL


class SerializerBase(AbstractSerializerBase):
    """
    serializer base class.

    all application serializers must inherit from this.
    """

    def __init__(self, *args, **options):
        """
        initializes an instance of SerializerBase.

        :param object args: constructor arguments.
        """

        super().__init__()

    def serialize(self, value, **options):
        """
        serializes the given value.

        returns serialized value on success or `NULL` object if serialization fails.

        :param object value: value to be serialized.

        :keyword ResultSchema result_schema: result schema instance to be
                                             used to create computed columns.
                                             defaults to None if not provided.

        :returns: serialized object
        :rtype: dict | list[dict]
        """

        if self.is_serializable(value, **options) is False:
            return NULL

        return self._serialize(value, **options)

    @abstractmethod
    def _serialize(self, value, **options):
        """
        serializes the given value.

        this method must be implemented by subclasses.

        :param object value: value to be serialized.

        :keyword ResultSchema result_schema: result schema instance to be
                                             used to create computed columns.
                                             defaults to None if not provided.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: serialized object
        :rtype: dict | list[dict]
        """

        raise CoreNotImplementedError()

    def is_serializable(self, value, **options):
        """
        gets a value indicating that the given input is serializable.

        :param object value: value to be serialized.

        :rtype: bool
        """

        return isinstance(value, self.accepted_type)



# -*- coding: utf-8 -*-
"""
city models module.
"""

from sqlalchemy import Unicode, Integer, ForeignKey, CheckConstraint

from pyrin.database.model.base import CoreEntity
from pyrin.database.orm.sql.schema.base import CoreColumn
from pyrin.utils.sqlalchemy import check_constraint


class CityBaseEntity(CoreEntity):
    """
    city base entity class.
    """

    _table = 'city'

    id = CoreColumn(name='id', autoincrement=True, type_=Integer, primary_key=True)


class CityEntity(CityBaseEntity):
    """
    city entity class.
    """

    _extend_existing = True

    @classmethod
    def _customize_table_args(cls, table_args):
        """
        customizes different table args for current entity type.

        this method is intended to be overridden by subclasses to customize
        table args per entity type if the required customization needs extra work.
        it must modify input dict values in-place if required.
        if other table args must be added (ex. UniqueConstraint or CheckConstraint ...)
        it must return those as a tuple. it could also return a single object as
        extra table arg (ex. a single UniqueConstraint).
        if no changes are required this method should not return anything.

        :param dict table_args: a dict containing different table args.
                                any changes to this dict must be done in-place.

        :rtype: tuple | object
        """

        return CheckConstraint(cls.name.notin_(['ali', 'reza']))
        # return check_constraint('name', ['ali', 'reza'], use_in=False)

    name = CoreColumn(name='name', type_=Unicode)

    province_id = CoreColumn(ForeignKey('province.id'),
                             name='province_id', type_=Integer)


class MyEntity(CoreEntity):
    """
    city base entity class.
    """

    _table = 'my_entity_table'

    id = CoreColumn(name='id', autoincrement=True, type_=Integer, primary_key=True)
    name = CoreColumn(name='name', type_=Unicode)
    level = CoreColumn(name='level', type_=Unicode)



# -*- coding: utf-8 -*-
"""
accounting services module.
"""

from pyrin.application.services import get_component

from demo.accounting import AccountingPackage


# Usage:
# you could implement different services here and call corresponding manager method this way:
# return get_component(AccountingPackage.COMPONENT_NAME).method_name(*arg, **kwargs)



# -*- coding: utf-8 -*-
"""
router component module.
"""

from pyrin.api.router import RouterPackage
from pyrin.api.router.manager import RouterManager
from pyrin.application.structs import Component
from pyrin.application.decorators import component


@component(RouterPackage.COMPONENT_NAME)
class RouterComponent(Component, RouterManager):
    """
    router component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
celery cli manager module.
"""

from celery.bin.celery import CeleryCommand

import pyrin.task_queues.celery.services as celery_services

from pyrin.cli.mixin.handler import CLIMixin
from pyrin.core.structs import Manager
from pyrin.task_queues.celery.cli import CeleryCLIPackage
from pyrin.task_queues.celery.cli.interface import CeleryCLIHandlerBase


class CeleryCLIManager(Manager, CLIMixin):
    """
    celery cli manager class.
    """

    _cli_handler_type = CeleryCLIHandlerBase
    package_class = CeleryCLIPackage

    def __init__(self):
        """
        initializes an instance of CeleryCLIManager.
        """

        super().__init__()
        self._command_executor = CeleryCommand(app=celery_services.get_current_app())

    def execute_on_celery(self, name, arguments=None):
        """
        executes the given celery command.

        :param str name: celery command name to be executed.
        :param list arguments: all command line arguments.

        :returns: result of celery command execution.
        """

        return self._command_executor.execute(name, arguments)



# -*- coding: utf-8 -*-
"""
response wrappers package.
"""

from pyrin.packaging.base import Package


class ResponseWrappersPackage(Package):
    """
    response wrappers package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
database exceptions module.
"""

from pyrin.core.exceptions import CoreException


class DatabaseManagerException(CoreException):
    """
    database manager exception.
    """
    pass


class InvalidSessionFactoryTypeError(DatabaseManagerException):
    """
    invalid session factory type error.
    """
    pass


class DuplicatedSessionFactoryError(DatabaseManagerException):
    """
    duplicated session factory error.
    """
    pass


class SessionFactoryNotExistedError(DatabaseManagerException):
    """
    session factory not existed error.
    """
    pass


class InvalidEntityTypeError(DatabaseManagerException):
    """
    invalid entity type error.
    """
    pass


class InvalidDatabaseBindError(DatabaseManagerException):
    """
    invalid database bind error.
    """
    pass


class InvalidDatabaseHookTypeError(DatabaseManagerException):
    """
    invalid database hook type error.
    """
    pass



# -*- coding: utf-8 -*-
"""
template enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class TemplateCLIHandlersEnum(CoreEnum):
    """
    template cli handlers enum.
    """

    PACKAGE = 'package'
    EMPTY_PACKAGE = 'empty_package'



# -*- coding: utf-8 -*-
"""
logging enumerations module.
"""

import logging

from pyrin.core.enumerations import CoreEnum


class LogLevelEnum(CoreEnum):
    """
    log level enum.
    """

    NOTSET = 'NOTSET'
    CRITICAL = 'CRITICAL'
    ERROR = 'ERROR'
    WARNING = 'WARNING'
    INFO = 'INFO'
    DEBUG = 'DEBUG'


class LogLevelIntEnum(CoreEnum):
    """
    log level int enum.
    """

    NOTSET = logging.NOTSET
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG



# -*- coding: utf-8 -*-
"""
alembic package.
"""

from pyrin.packaging.base import Package


class AlembicPackage(Package):
    """
    alembic package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'database.migration.alembic.component'
    CONFIG_STORE_NAMES = ['alembic']
    DEPENDS = ['pyrin.configuration',
               'pyrin.cli']



# -*- coding: utf-8 -*-
"""
logging services module.
"""

from pyrin.application.services import get_component
from pyrin.logging import LoggingPackage


def get_all_loggers():
    """
    gets a dictionary containing all available loggers.

    it returns a shallow copy of loggers dict.

    :returns: dict[str name: Logger instance]
    :rtype: dict
    """

    return get_component(LoggingPackage.COMPONENT_NAME).get_all_loggers()


def should_be_wrapped(logger):
    """
    gets a value indication that given logger should be wrapped.

    note that we should not wrap sqlalchemy and alembic loggers,
    because it does not affect on sqlalchemy loggers and they
    never have request info in emitted logs, I don't know the reason.
    but wrapping them actually has a side effect which leads to some errors.
    so we do not wrap them in the first place.

    :param Logger logger: logger to check should it be wrapped.

    :rtype: bool
    """

    return get_component(LoggingPackage.COMPONENT_NAME).should_be_wrapped(logger)


def reload_configs(**options):
    """
    reloads all logging configurations from config file.
    """

    return get_component(LoggingPackage.COMPONENT_NAME).reload_configs(**options)


def wrap_all_loggers():
    """
    wraps all available loggers into an adapter.

    normally, this method should not be called manually.
    """

    get_component(LoggingPackage.COMPONENT_NAME).wrap_all_loggers()


def get_logger(name, **options):
    """
    gets the logger based on input parameters.

    :param str name: logger name to get.

    :returns: specified logger.
    :rtype: Logger
    """

    return get_component(LoggingPackage.COMPONENT_NAME).get_logger(name, **options)


def debug(msg, *args, **kwargs):
    """
    emits a log with debug level.

    :param str msg: log message.

    :keyword dict interpolation_data: data to be used for interpolation.
    """

    return get_component(LoggingPackage.COMPONENT_NAME).debug(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    """
    emits a log with info level.

    :param str msg: log message.

    :keyword dict interpolation_data: data to be used for interpolation.
    """

    return get_component(LoggingPackage.COMPONENT_NAME).info(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):
    """
    emits a log with warning level.

    :param str msg: log message.

    :keyword dict interpolation_data: data to be used for interpolation.
    """

    return get_component(LoggingPackage.COMPONENT_NAME).warning(msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    """
    emits a log with error level.

    :param str msg: log message.

    :keyword dict interpolation_data: data to be used for interpolation.
    """

    return get_component(LoggingPackage.COMPONENT_NAME).error(msg, *args, **kwargs)


def exception(msg, *args, **kwargs):
    """
    emits a log with error level and exception information.

    :param str msg: log message.

    :keyword dict interpolation_data: data to be used for interpolation.
    """

    return get_component(LoggingPackage.COMPONENT_NAME).exception(msg, *args, **kwargs)


def critical(msg, *args, **kwargs):
    """
    emits a log with critical level.

    :param str msg: log message.

    :keyword dict interpolation_data: data to be used for interpolation.
    """

    return get_component(LoggingPackage.COMPONENT_NAME).critical(msg, *args, **kwargs)


def register_hook(instance):
    """
    registers the given instance into logging hooks.

    :param LoggingHookBase instance: logging hook instance to be registered.

    :raises InvalidLoggingHookTypeError: invalid logging hook type error.
    """

    return get_component(LoggingPackage.COMPONENT_NAME).register_hook(instance)


def prepare_data(data, **options):
    """
    this method will call `prepare_data` method of all registered hooks.

    :param dict | object data: data that is passed to logging method.

    :returns: modified or same input data.
    :rtype: dict | object
    """

    return get_component(LoggingPackage.COMPONENT_NAME).prepare_data(data, **options)


def before_emit(message, data, level, **options):
    """
    this method will call `before_emit` method of all registered hooks.

    :param str message: the log message that must be emitted.
    :param dict | object data: data that is passed to logging method.
    :param int level: log level.
    """

    return get_component(LoggingPackage.COMPONENT_NAME).before_emit(message, data,
                                                                    level, **options)


def after_emit(message, data, level, **options):
    """
    this method will call `after_emit` method of all registered hooks.

    :param str message: the log message that has been emitted.
    :param dict | object data: data that is passed to logging method.
    :param int level: log level.
    """

    return get_component(LoggingPackage.COMPONENT_NAME).after_emit(message, data,
                                                                   level, **options)



# -*- coding: utf-8 -*-
"""
alembic services module.
"""

from pyrin.application.services import get_component
from pyrin.database.migration.alembic import AlembicPackage


def register_cli_handler(instance, **options):
    """
    registers a new alembic cli handler or replaces the existing one
    if `replace=True` is provided. otherwise, it raises an error
    on adding a cli handler which is already registered.

    :param AlembicCLIHandlerBase instance: alembic cli handler to be registered.
                                           it must be an instance of
                                           AlembicCLIHandlerBase.

    :keyword bool replace: specifies that if there is another registered
                           cli handler with the same name, replace it
                           with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidCLIHandlerTypeError: invalid cli handler type error.
    :raises DuplicatedCLIHandlerError: duplicated cli handler error.
    """

    get_component(AlembicPackage.COMPONENT_NAME).register_cli_handler(instance, **options)


def execute(handler_name, **options):
    """
    executes the handler with the given name with given inputs.

    :param str handler_name: handler name to be executed.

    :raises CLIHandlerNotFoundError: cli handler not found error.
    """

    return get_component(AlembicPackage.COMPONENT_NAME).execute(handler_name, **options)


def enable():
    """
    enables migrations for the application.
    """

    return get_component(AlembicPackage.COMPONENT_NAME).enable()


def get_package_class():
    """
    gets the package class of alembic manager.

    :raises PackageClassIsNotSetError: package class is not set error.

    :returns: type[AlembicPackage]
    """

    return get_component(AlembicPackage.COMPONENT_NAME).get_package_class()



# -*- coding: utf-8 -*-
"""
caching remote handlers base module.
"""

from abc import abstractmethod

import pyrin.globalization.datetime.services as datetime_services
import pyrin.configuration.services as config_services
import pyrin.logging.services as logging_services

from pyrin.caching.interface import AbstractRemoteCache
from pyrin.caching.mixin.base import ComplexKeyGeneratorMixin
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.caching.exceptions import CacheNameIsRequiredError, InvalidCacheExpireTimeError


class RemoteCacheBase(ComplexKeyGeneratorMixin, AbstractRemoteCache):
    """
    remote cache base class.

    it could be used as the base class for all remote caches, such as memcached and redis.
    """

    # cache name to be used for this cache.
    # it must be unique between all caches.
    cache_name = None

    LOGGER = logging_services.get_logger('caching.remote')

    def __init__(self, *args, **options):
        """
        initializes an instance of RemoteCacheBase.

        :keyword int expire: default expire time of cached items in milliseconds.
                             if not provided, it will be get from `caching` config
                             store.

        :keyword bool consider_user: specifies that current user must also be
                                     included in cache key. if not provided, will
                                     be get from `caching` config store.

        :raises CacheNameIsRequiredError: cache name is required error.
        :raises InvalidCacheExpireTimeError: invalid cache expire time error.
        """

        super().__init__()

        if self.cache_name in (None, '') or self.cache_name.isspace():
            raise CacheNameIsRequiredError('Cache name must be provided for '
                                           'cache [{name}].'.format(name=self))

        self._set_name(self.cache_name)

        expire = options.get('expire')
        consider_user = options.get('consider_user')

        configs = self._get_configs()
        if expire is None:
            expire = configs.pop('expire')
        configs.pop('expire', None)

        if consider_user is None:
            consider_user = configs.pop('consider_user')
        configs.pop('consider_user', None)

        if expire < 0:
            raise InvalidCacheExpireTimeError('Cache expire time for cache [{name}] '
                                              'must be a non-negative integer.'
                                              .format(name=self.get_name()))

        self._last_cleared_time = datetime_services.now()
        self._expire = expire
        self._consider_user = consider_user
        self._client = self._create_client(*args, kwargs=options, **configs)

    @abstractmethod
    def _create_client(self, *args, kwargs=None, **configs):
        """
        creates a client for connecting to remote cache server.

        :param object args: all positional arguments passed to `__init__` method.
        :param dict kwargs: all keyword arguments passed to `__init__` method.

        :keyword object configs: all configurations of this cache from `caching`
                                 config store.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _clear(self):
        """
        clears all items from cache.

        this method is intended to be overridden in subclasses.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _set(self, key, value, *args, **options):
        """
        sets a new value into cached items.

        this method must be overridden in subclasses.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get(self, key, default=None, **options):
        """
        gets the value from cache.

        if key does not exist, it returns None or the specified default value.
        this method must be overridden in subclasses.

        :param object key: hashable key to get its value from cache.
        :param object default: value to be returned if key is not present.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    def _prepare_key(self, key):
        """
        prepares the key to be used in cache.

        this method is intended to be overridden in subclasses.

        :param object key: key to be cached.

        :returns: prepared key.
        """

        return key

    def _prepare_for_set(self, value):
        """
        prepares the value that must be cached.

        this method is intended to be overridden in subclasses.

        :param object value: value to be cached.

        :returns: prepared value.
        """

        return value

    def _prepare_for_get(self, value):
        """
        prepares the value that must be returned from cache.

        this method is intended to be overridden in subclasses.

        :param object value: value to be returned.

        :returns: prepared value.
        """

        return value

    def _get_configs(self):
        """
        gets the configs of this cache from `caching` config store.

        :rtype: dict
        """

        return config_services.get_section('caching', self.get_name())

    def _get_hit_ratio(self):
        """
        gets hit ratio for this cache in percentage.

        :rtype: float
        """

        hit = self.hit_count
        miss = self.miss_count
        if hit == 0 and miss == 0:
            return 0

        ratio = hit / (hit + miss)
        return ratio * 100

    def _try_set(self, value, func, inputs, kw_inputs, *args, **options):
        """
        sets a new value into cached items.

        this method will generate cache key from given type and function
        and inputs. it also considers current component key in key generation.

        :param object value: value to be cached.
        :param function func: function to cache its result.
        :param tuple inputs: function positional arguments.
        :param dict kw_inputs: function keyword arguments.

        :keyword bool consider_user: specifies that current user must be included in
                                     key generation. it will be get from `consider_user`
                                     attribute store if not provided.
        """

        key = self.generate_key(func, inputs, kw_inputs, *args, **options)
        self.set(key, value, **options)

    def _try_get(self, func, inputs, kw_inputs, *args, default=None, **options):
        """
        gets the value from cache.

        this method will generate cache key from given type and function
        and inputs. it also considers current component key in key generation.
        if key does not exist, it returns None or the specified default value.

        :param function func: function to to get its result.
        :param tuple inputs: function positional arguments.
        :param dict kw_inputs: function keyword arguments.
        :param object default: value to be returned if key is not present.

        :keyword bool consider_user: specifies that current user must be included in
                                     key generation. it will be get from `consider_user`
                                     attribute store if not provided.

        :returns: object
        """

        key = self.generate_key(func, inputs, kw_inputs, *args, **options)
        return self.get(key, default, **options)

    def set(self, key, value, *args, **options):
        """
        sets a new value into cached items.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)
        prepared_value = self._prepare_for_set(value)
        self._set(hashed_key, prepared_value, *args, **options)

    def get(self, key, default=None, **options):
        """
        gets the value from cache.

        if key does not exist, it returns None or the specified default value.

        :param object key: hashable key to get its value from cache.
        :param object default: value to be returned if key is not present.

        :returns: object
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)
        result = self._get(hashed_key, default=default, **options)
        if result in (None, default):
            return result

        return self._prepare_for_get(result)

    def try_set(self, value, func, inputs, kw_inputs, *args, **options):
        """
        sets a new value into cached items.

        this method will generate cache key from given type and function
        and inputs. it also considers current component key in key generation.

        :param object value: value to be cached.
        :param function func: function to cache its result.
        :param tuple inputs: function positional arguments.
        :param dict kw_inputs: function keyword arguments.

        :keyword bool consider_user: specifies that current user must be included in
                                     key generation. it will be get from `consider_user`
                                     attribute store if not provided.
        """

        try:
            self._try_set(value, func, inputs, kw_inputs, *args, **options)
        except TypeError as error:
            self.LOGGER.exception(str(error))

    def try_get(self, func, inputs, kw_inputs, *args, default=None, **options):
        """
        gets the value from cache.

        this method will generate cache key from given type and function
        and inputs. it also considers current component key in key generation.
        if key does not exist, it returns None or the specified default value.

        :param function func: function to to get its result.
        :param tuple inputs: function positional arguments.
        :param dict kw_inputs: function keyword arguments.
        :param object default: value to be returned if key is not present.

        :keyword bool consider_user: specifies that current user must be included in
                                     key generation. it will be get from `consider_user`
                                     attribute store if not provided.

        :returns: object
        """

        try:
            return self._try_get(func, inputs, kw_inputs, *args, default=default, **options)
        except TypeError as error:
            self.LOGGER.exception(str(error))
            return None

    def clear(self):
        """
        clears all items from cache.
        """

        self._clear()
        self._last_cleared_time = datetime_services.now()

    @property
    def consider_user(self):
        """
        gets the consider user value for this cache.

        :rtype: bool
        """

        return self._consider_user

    @property
    def last_cleared_time(self):
        """
        gets the last time in which this cache has been cleared.

        :rtype: datetime.datetime
        """

        return self._last_cleared_time

    @property
    def stats(self):
        """
        get the statistic info about cached items.

        :returns: dict(datetime last_cleared_time: last cleared time,
                       bool persistent: persistent cache,
                       bool consider_user: consider user,
                       int expire: cached items expire time,
                       int hit: hit count,
                       int miss: miss count,
                       str hit_ratio: hit ratio)
        :rtype: dict
        """

        hit_ratio = self._get_hit_ratio()
        hit_ratio = '{:0.1f}%'.format(hit_ratio)

        return dict(last_cleared_time=self.last_cleared_time,
                    persistent=self.persistent,
                    consider_user=self.consider_user,
                    expire=self.expire,
                    hit=self.hit_count,
                    miss=self.miss_count,
                    hit_ratio=hit_ratio)

    @property
    def persistent(self):
        """
        gets a value indicating that cached items must be persisted to database on shutdown.

        :rtype: bool
        """

        return False

    @property
    def expire(self):
        """
        gets default expire time value for this cache's items in milliseconds.

        :rtype: int
        """

        return self._expire

    @property
    def client(self):
        """
        gets the client of this cache.

        :returns: object
        """

        return self._client



# -*- coding: utf-8 -*-
"""
caching mixin decorators module.
"""

from functools import update_wrapper


def fast_cache(*old_method):
    """
    decorator to convert a method or property into a lazy one.

    note that this type of caches will generate cache key based on
    parent type and function name. so it could only be used on instance or class
    level methods and properties. the cached method also could not have any arguments.

    note that this decorator should only be used in classes which subclassed from
    `TypedCacheMixin` class.

    this type of caches are implemented to be used where performance is critical.
    it has as few overhead as possible even on single usage.

    :param function | property old_method: the original decorated method or property.

    :returns: method or property result.
    """

    def decorator(method):
        """
        decorates the given method or property and makes it a lazy one.

        :param function | property method: decorated method or property.

        :returns: method or property result.
        """

        def wrapper(self):
            """
            decorates the given method or property and makes it a lazy one.

            :returns: method or property result.
            """

            result = self.get_cache(method)
            if result is not None:
                return result

            result = method(self)
            self.set_cache(result, method)

            return result

        return update_wrapper(wrapper, method)

    if len(old_method) > 0:
        return decorator(old_method[0])

    return decorator


def extended_fast_cache(*old_method, **options):
    """
    decorator to convert a method or property into a lazy one.

    note that this type of caches will generate cache key based on
    parent type, function name, method inputs, current user and component key.
    so it could only be used on instance or class level methods and properties.

    note that this decorator should only be used in classes which subclassed from
    `ExtendedTypedCacheMixin` class.

    this type of caches are implemented to be used where performance is critical.
    it has as few overhead as possible but it is slower than `fast_cache` because
    it processes function inputs.

    important note: if you want to use this decorator inside pyrin code, you must
    add `pyrin.security.session` into depends list of the package that uses this
    decorator. this is not required in top level application.

    :param function | property old_method: the original decorated method or property.

    :keyword bool consider_user: specifies that current user must be included in
                                 key generation. defaults to True if not provided.

    :returns: method or property result.
    """

    def decorator(method):
        """
        decorates the given method or property and makes it a lazy one.

        :param function | property method: decorated method or property.

        :returns: method or property result.
        """

        def wrapper(self, *args, **kwargs):
            """
            decorates the given method or property and makes it a lazy one.

            :param object args: function positional arguments.
            :param object kwargs: function keyword arguments.

            :returns: method or property result.
            """

            result = self.get_cache(method, args, kwargs, **options)
            if result is not None:
                return result

            result = method(self, *args, **kwargs)
            self.set_cache(result, method, args, kwargs, **options)

            return result

        return update_wrapper(wrapper, method)

    if len(old_method) > 0:
        return decorator(old_method[0])

    return decorator



# -*- coding: utf-8 -*-
"""
cors manager module.
"""

from werkzeug.exceptions import HTTPException

import pyrin.application.services as application_services
import pyrin.security.session.services as session_services
import pyrin.configuration.services as config_services
import pyrin.utils.string as string_utils

from pyrin.processor.cors import CORSPackage
from pyrin.core.structs import Manager, CoreHeaders
from pyrin.processor.cors.enumerations import CORSResponseHeaderEnum, CORSRequestHeaderEnum
from pyrin.processor.request.enumerations import RequestHeaderEnum
from pyrin.processor.response.enumerations import ResponseHeaderEnum


class CORSManager(Manager):
    """
    cors manager class.
    """

    package_class = CORSPackage
    WILDCARD = '*'

    def __init__(self):
        """
        initializes an instance of CORSManager.
        """

        super().__init__()

        self._enabled = config_services.get_active('cors', 'enable')
        self._always_send = config_services.get_active('cors', 'always_send')
        self._allowed_origins = set(config_services.get_active('cors', 'allowed_origins') or [])
        self._allowed_headers = set(config_services.get_active('cors', 'allowed_headers') or [])
        self._exposed_headers = set(config_services.get_active('cors', 'exposed_headers') or [])
        self._allow_credentials = config_services.get_active('cors', 'allow_credentials')
        self._max_age = config_services.get_active('cors', 'max_age')

    def _validate_credentials(self, items, **options):
        """
        validates that if credentials are supported or not.

        if credentials are supported, it removes `*` from values in the given list.

        :param list[str] | set[str] items: items to be validated.

        :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                         response headers to front-end javascript code
                                         if the route is authenticated.
                                         if not provided, defaults to `_allow_credentials`
                                         attribute value.
        """

        if self._should_allow_credentials(**options) is not False:
            if self.WILDCARD in items:
                items.remove(self.WILDCARD)

    def _is_enabled(self, **options):
        """
        gets a value indicating that cors headers must be sent in response.

        :keyword bool enabled: specifies that cors headers must be sent in response.
                               if not provided, defaults to `_enabled` attribute value.

        :rtype: bool
        """

        enabled = options.get('enabled')
        if enabled is None:
            enabled = self._enabled

        return enabled

    def _should_always_send(self, **options):
        """
        gets a value indicating that cors headers must be included in response
        even if the request does not have origin header.

        :keyword bool always_send: specifies that cors headers must be included in
                                   response even if the request does not have origin header.
                                   if not provided, defaults to `_always_send` attribute value.
        :rtype: bool
        """

        always_send = options.get('always_send')
        if always_send is None:
            always_send = self._always_send

        return always_send

    def _should_allow_credentials(self, **options):
        """
        gets a value indicating that browsers are allowed to pass response
        headers to front-end javascript code if the route is authenticated.

        :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                         response headers to front-end javascript code
                                         if the route is authenticated.
                                         if not provided, defaults to `_allow_credentials`
                                         attribute value.

        :rtype: bool
        """

        allow_credentials = options.get('allow_credentials')
        if allow_credentials is None:
            allow_credentials = self._allow_credentials

        return allow_credentials

    def _get_allowed_origins(self, **options):
        """
        gets all allowed origins combining default and extra ones.

        :keyword list[str] allowed_origins: extra allowed origins to be combined
                                            with default ones.

        :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                         response headers to front-end javascript code
                                         if the route is authenticated.
                                         if not provided, defaults to `_allow_credentials`
                                         attribute value.

        :rtype: set[str]
        """

        allowed_origins = options.get('allowed_origins') or []
        allowed_origins = set(allowed_origins).union(self._allowed_origins)
        self._validate_credentials(allowed_origins, **options)
        return allowed_origins

    def _get_allowed_headers(self, **options):
        """
        gets all allowed request headers combining default and extra ones.

        :keyword list[str] allowed_headers: extra allowed headers to be combined
                                            with default ones.

        :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                         response headers to front-end javascript code
                                         if the route is authenticated.
                                         if not provided, defaults to `_allow_credentials`
                                         attribute value.

        :rtype: set[str]
        """

        allowed_headers = options.get('allowed_headers') or []
        allowed_headers = set(allowed_headers).union(self._allowed_headers)
        self._validate_credentials(allowed_headers, **options)
        return allowed_headers

    def _get_exposed_headers(self, **options):
        """
        gets all exposed response headers combining default and extra ones.

        :keyword list[str] exposed_headers: extra exposed headers to be combined
                                            with default ones.

        :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                         response headers to front-end javascript code
                                         if the route is authenticated.
                                         if not provided, defaults to `_allow_credentials`
                                         attribute value.

        :rtype: list[str]
        """

        exposed_headers = options.get('exposed_headers') or []
        exposed_headers = sorted(set(exposed_headers).union(self._exposed_headers))
        self._validate_credentials(exposed_headers, **options)
        return exposed_headers

    def _get_max_age(self, **options):
        """
        gets the max age for cors enabled routes to be cached in seconds.

        it returns None if no max age is set.

        :keyword int max_age: maximum number of seconds to cache results.
                              if not provided, defaults to `_max_age` attribute value.

        :rtype: int
        """

        return options.setdefault('max_age', self._max_age)

    def _get_response_origin(self, origin, **options):
        """
        gets the origin to be set in response header.

        it returns None if conditions are not met.

        :param str origin: the origin name from request.

        :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                         response headers to front-end javascript code
                                         if the route is authenticated.
                                         if not provided, defaults to `_allow_credentials`
                                         attribute value.

        :keyword list[str] allowed_origins: a list of extra allowed origins to be used
                                            in conjunction with default allowed ones.

        :rtype: str
        """

        allowed_origins = self._get_allowed_origins(**options)
        if origin not in (None, '', self.WILDCARD):
            if string_utils.is_match(origin, allowed_origins) is True:
                return origin

            if self.WILDCARD in allowed_origins:
                return self.WILDCARD

        elif self._should_always_send(**options) is True and \
                self.WILDCARD in allowed_origins:
            return self.WILDCARD

        return None

    def _get_response_allowed_headers(self, headers, **options):
        """
        gets those requested headers which are allowed for cors enabled routes.

        :param list[str] headers: list of requested headers to be checked if are allowed.

        :keyword list[str] allowed_headers: extra allowed headers to be combined
                                            with default ones.

        :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                         response headers to front-end javascript code
                                         if the route is authenticated.
                                         if not provided, defaults to `_allow_credentials`
                                         attribute value.

        :rtype: list[str]
        """

        allowed_headers = self._get_allowed_headers(**options)
        matching_headers = string_utils.union(headers, allowed_headers, collection=set)
        if self.WILDCARD in allowed_headers:
            matching_headers.add(self.WILDCARD)

        return sorted(matching_headers)

    def _get_common_headers(self, **options):
        """
        gets all headers that are common between preflight and actual requests.

        it returns None if cors is not enabled or request's origin is not valid.

        :keyword bool enabled: specifies that cors headers must be sent in response.
                               if not provided, defaults to `_enabled` attribute value.

        :keyword bool always_send: specifies that cors headers must be included in
                                   response even if the request does not have origin header.
                                   if not provided, defaults to `_always_send` attribute value.

        :keyword list[str] allowed_origins: a list of extra allowed origins to be used
                                            in conjunction with default allowed ones.

        :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                         response headers to front-end javascript code
                                         if the route is authenticated.
                                         if not provided, defaults to `_allow_credentials`
                                         attribute value.

        :rtype: CoreHeaders
        """

        if self._is_enabled(**options) is not True:
            return None

        request = session_services.get_current_request()
        origin = self._get_response_origin(request.origin, **options)

        # if origin is not provided in request header and also wildcard
        # condition is not met, cors headers will not be added.
        if origin is None:
            return None

        headers = CoreHeaders()
        headers[CORSResponseHeaderEnum.ACCESS_CONTROL_ALLOW_ORIGIN] = origin

        vary_headers = [CORSRequestHeaderEnum.ACCESS_CONTROL_REQUEST_METHOD]
        if origin != self.WILDCARD:
            vary_headers.append(RequestHeaderEnum.ORIGIN)

        vary_headers = ', '.join(vary_headers)
        headers.add(ResponseHeaderEnum.VARY, vary_headers)

        if self._should_allow_credentials(**options) is True:
            headers[CORSResponseHeaderEnum.ACCESS_CONTROL_ALLOW_CREDENTIALS] = 'true'

        return headers

    def get_cors_headers(self, **options):
        """
        gets all headers to set in response for cors enabled actual requests.

        it returns None if cors is not enabled or request's origin is not valid.

        :keyword bool enabled: specifies that cors headers must be sent in response.
                               if not provided, defaults to `_enabled` attribute value.

        :keyword bool always_send: specifies that cors headers must be included in
                                   response even if the request does not have origin header.
                                   if not provided, defaults to `_always_send` attribute value.

        :keyword list[str] allowed_origins: a list of extra allowed origins to be used
                                            in conjunction with default allowed ones.

        :keyword list[str] exposed_headers: extra exposed headers to be combined
                                            with default ones.

        :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                         response headers to front-end javascript code
                                         if the route is authenticated.
                                         if not provided, defaults to `_allow_credentials`
                                         attribute value.

        :rtype: CoreHeaders
        """

        headers = self._get_common_headers(**options)
        if headers is None:
            return None

        exposed_headers = self._get_exposed_headers(**options)
        if len(exposed_headers) > 0:
            exposed_headers = ', '.join(exposed_headers)
            headers[CORSResponseHeaderEnum.ACCESS_CONTROL_EXPOSE_HEADERS] = exposed_headers

        return headers

    def get_preflight_headers(self, *allowed_methods, **options):
        """
        gets all headers to set in response for cors preflight requests.

        it returns None if cors is not enabled or request's origin is not valid.

        :param str allowed_methods: all allowed http methods.

        :keyword bool enabled: specifies that cors headers must be sent in response.
                               if not provided, defaults to `_enabled` attribute value.

        :keyword bool always_send: specifies that cors headers must be included in
                                   response even if the request does not have origin header.
                                   if not provided, defaults to `_always_send` attribute value.

        :keyword list[str] allowed_origins: a list of extra allowed origins to be used
                                            in conjunction with default allowed ones.

        :keyword list[str] allowed_headers: extra allowed headers to be combined
                                            with default ones.

        :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                         response headers to front-end javascript code
                                         if the route is authenticated.
                                         if not provided, defaults to `_allow_credentials`
                                         attribute value.

        :keyword int max_age: maximum number of seconds to cache results.
                              if not provided, defaults to `_max_age` attribute value.

        :rtype: CoreHeaders
        """

        headers = self._get_common_headers(**options)
        if headers is None:
            return None

        # if there is no 'Access-Control-Request-Method' in request headers or if it
        # does not match any of allowed methods, do not add other preflight headers.
        request = session_services.get_current_request()
        if request.access_control_request_method in (None, '') or \
                string_utils.is_match(request.access_control_request_method,
                                      allowed_methods) is False:
            return headers

        allowed_methods = ', '.join(sorted([item.upper() for item in allowed_methods]))
        headers[CORSResponseHeaderEnum.ACCESS_CONTROL_ALLOW_METHODS] = allowed_methods

        requested_headers = []
        if request.access_control_request_headers not in (None, ''):
            requested_headers = \
                request.access_control_request_headers.as_set(preserve_casing=True)

        allowed_headers = self._get_response_allowed_headers(requested_headers, **options)
        if len(allowed_headers) > 0:
            allowed_headers = ', '.join(allowed_headers)
            headers[CORSResponseHeaderEnum.ACCESS_CONTROL_ALLOW_HEADERS] = allowed_headers

        max_age = self._get_max_age(**options)
        if max_age is not None:
            headers[CORSResponseHeaderEnum.ACCESS_CONTROL_MAX_AGE] = max_age

        return headers

    def process_inputs(self, cors, **options):
        """
        processes given cors object and gets the relevant inputs as a dict.

        if any of cors objects attributes are None, it gets their values from default configs.

        :param CORS cors: cors object.

        :returns: dict(bool enabled,
                       bool always_send,
                       list[str] allowed_origins,
                       list[str] allowed_headers,
                       list[str] exposed_headers,
                       bool allow_credentials,
                       int max_age)

        :rtype: dict
        """

        enabled = cors.enabled
        always_send = cors.always_send
        allowed_origins = cors.allowed_origins
        allowed_headers = cors.allowed_headers
        exposed_headers = cors.exposed_headers
        allow_credentials = cors.allow_credentials
        max_age = cors.max_age

        if enabled is None:
            enabled = self._enabled

        if always_send is None:
            always_send = self._always_send

        if allowed_origins is None:
            allowed_origins = self._allowed_origins

        if allowed_headers is None:
            allowed_headers = self._allowed_headers

        if exposed_headers is None:
            exposed_headers = self._exposed_headers

        if allow_credentials is None:
            allow_credentials = self._allow_credentials

        if max_age is None:
            max_age = self._max_age

        return dict(enabled=enabled,
                    always_send=always_send,
                    allowed_origins=allowed_origins,
                    allowed_headers=allowed_headers,
                    exposed_headers=exposed_headers,
                    allow_credentials=allow_credentials,
                    max_age=max_age)

    def get_required_cors_headers(self):
        """
        gets cors headers for current request if required.

        if cors conditions are not met, it returns None.

        :rtype: CoreHeaders
        """

        request = session_services.get_current_request()
        if request.url_rule is not None:
            inputs = self.process_inputs(request.url_rule.cors)
            return self.get_cors_headers(**inputs)

        return None

    def get_required_preflight_headers(self):
        """
        gets preflight headers for current request if required.

        if any errors occurs or cors conditions are not met, it returns None.

        :rtype: CoreHeaders
        """

        request = session_services.get_current_request()
        adapter = application_services.get_current_url_adapter()
        headers = None
        try:
            rule, arguments = adapter.match(method=request.access_control_request_method,
                                            return_rule=True)

            inputs = self.process_inputs(rule.cors)
            headers = self.get_preflight_headers(request.access_control_request_method,
                                                 **inputs)

        except HTTPException:
            pass

        return headers

    def get_current_cors_headers(self):
        """
        gets all required cors or preflight headers.

        it may return None if cors conditions are not met.

        :rtype: CoreHeaders
        """

        request = session_services.get_current_request()
        if request.is_preflight is True:
            return self.get_required_preflight_headers()

        elif request.is_cors is True:
            return self.get_required_cors_headers()

        return None



# -*- coding: utf-8 -*-
"""
cors services module.
"""

from pyrin.processor.cors import CORSPackage
from pyrin.application.services import get_component


def get_cors_headers(**options):
    """
    gets all headers to set in response for cors enabled actual requests.

    it returns None if cors is not enabled or request's origin is not valid.

    :keyword bool enabled: specifies that cors headers must be sent in response.
                           if not provided, defaults to `_enabled` attribute value.

    :keyword bool always_send: specifies that cors headers must be included in
                               response even if the request does not have origin header.
                               if not provided, defaults to `_always_send` attribute value.

    :keyword list[str] allowed_origins: a list of extra allowed origins to be used
                                        in conjunction with default allowed ones.

    :keyword list[str] exposed_headers: extra exposed headers to be combined
                                        with default ones.

    :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                     response headers to front-end javascript code
                                     if the route is authenticated.
                                     if not provided, defaults to `_allow_credentials`
                                     attribute value.

    :rtype: CoreHeaders
    """

    return get_component(CORSPackage.COMPONENT_NAME).get_cors_headers(**options)


def get_preflight_headers(*allowed_methods, **options):
    """
    gets all headers to set in response for cors preflight requests.

    it returns None if cors is not enabled or request's origin is not valid.

    :param str allowed_methods: all allowed http methods.

    :keyword bool enabled: specifies that cors headers must be sent in response.
                           if not provided, defaults to `_enabled` attribute value.

    :keyword bool always_send: specifies that cors headers must be included in
                               response even if the request does not have origin header.
                               if not provided, defaults to `_always_send` attribute value.

    :keyword list[str] allowed_origins: a list of extra allowed origins to be used
                                        in conjunction with default allowed ones.

    :keyword list[str] allowed_headers: extra allowed headers to be combined
                                        with default ones.

    :keyword bool allow_credentials: specifies that browsers are allowed to pass
                                     response headers to front-end javascript code
                                     if the route is authenticated.
                                     if not provided, defaults to `_allow_credentials`
                                     attribute value.

    :keyword int max_age: maximum number of seconds to cache results.
                          if not provided, defaults to `_max_age` attribute value.

    :rtype: CoreHeaders
    """

    return get_component(CORSPackage.COMPONENT_NAME).get_preflight_headers(*allowed_methods,
                                                                           **options)


def process_inputs(cors, **options):
    """
    processes given cors object and gets the relevant inputs as a dict.

    if any of cors objects attributes are None, it gets their values from default configs.

    :param CORS cors: cors object.

    :returns: dict(bool enabled,
                   bool always_send,
                   list[str] allowed_origins,
                   list[str] allowed_headers,
                   list[str] exposed_headers,
                   bool allow_credentials,
                   int max_age)

    :rtype: dict
    """

    return get_component(CORSPackage.COMPONENT_NAME).process_inputs(cors, **options)


def get_required_cors_headers():
    """
    gets cors headers for current request if required.

    if cors conditions are not met, it returns None.

    :rtype: CoreHeaders
    """

    return get_component(CORSPackage.COMPONENT_NAME).get_required_cors_headers()


def get_required_preflight_headers():
    """
    gets preflight headers for current request if required.

    if any errors occurs or cors conditions are not met, it returns None.

    :rtype: CoreHeaders
    """

    return get_component(CORSPackage.COMPONENT_NAME).get_required_preflight_headers()


def get_current_cors_headers():
    """
    gets all required cors or preflight headers.

    it may return None if cors conditions are not met.

    :rtype: CoreHeaders
    """

    return get_component(CORSPackage.COMPONENT_NAME).get_current_cors_headers()



# -*- coding: utf-8 -*-
"""
database migration component module.
"""

from pyrin.application.decorators import component
from pyrin.database.migration import DatabaseMigrationPackage
from pyrin.database.migration.manager import DatabaseMigrationManager
from pyrin.application.structs import Component


@component(DatabaseMigrationPackage.COMPONENT_NAME)
class DatabaseMigrationComponent(Component, DatabaseMigrationManager):
    """
    database migration component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
deserializer handlers number module.
"""

import re

from pyrin.converters.deserializer.handlers.base import StringPatternDeserializerBase
from pyrin.converters.deserializer.decorators import deserializer


@deserializer()
class IntegerDeserializer(StringPatternDeserializerBase):
    """
    integer deserializer class.
    """

    # matches the integer inside string.
    # example: 12, 232, 10, 0, -5, +7
    # all of these values will be matched.
    INTEGER_REGEX = re.compile(r'^[+|\-]?([0-9])+$')

    def __init__(self, **options):
        """
        creates an instance of IntegerDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  formats and their min and
                                                                  max length for integer
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        :param str value: value to be deserialized.

        :keyword Pattern matching_pattern: the pattern that has matched the value.

        :rtype: int
        """

        return int(value)

    @property
    def default_formats(self):
        """
        gets default accepted formats that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(self.INTEGER_REGEX, self.UNDEF_LENGTH, self.UNDEF_LENGTH)]


@deserializer()
class FloatDeserializer(StringPatternDeserializerBase):
    """
    float deserializer class.
    """

    # default min for this deserializer is 3 because it
    # should at least have two digits and a dot between them.
    DEFAULT_MIN = 3

    # matches the float inside string.
    # example: 0.12, 2.32, 1.0, 0.0, -1.6, +5.06
    # all of these values will be matched.
    FLOAT_REGEX = re.compile(r'^[+|\-]?([0-9])+[.]([0-9])+$')

    def __init__(self, **options):
        """
        creates an instance of FloatDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  formats and their min and
                                                                  max length for float
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        :param str value: value to be deserialized.

        :rtype: float
        """

        return float(value)

    @property
    def default_formats(self):
        """
        gets default accepted formats that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(self.FLOAT_REGEX, self.UNDEF_LENGTH, self.UNDEF_LENGTH)]



# -*- coding: utf-8 -*-
"""
alembic handlers current module.
"""

from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.enumerations import AlembicCLIHandlersEnum
from pyrin.database.migration.alembic.handlers.base import AlembicReportingCLIHandlerBase


@alembic_cli_handler()
class CurrentCLIHandler(AlembicReportingCLIHandlerBase):
    """
    current cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of CurrentCLIHandler.
        """

        super().__init__(AlembicCLIHandlersEnum.CURRENT)



# -*- coding: utf-8 -*-
"""
users package.
"""

from pyrin.security.users import UsersPackage as BaseUsersPackage


class UsersPackage(BaseUsersPackage):
    """
    users package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
validator package.
"""

from pyrin.packaging.base import Package


class ValidatorPackage(Package):
    """
    validator package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'validator.component'



# -*- coding: utf-8 -*-
"""
encryption decorators module.
"""

import pyrin.security.encryption.services as encryption_services


def encrypter(*args, **kwargs):
    """
    decorator to register an encryption handler.

    :param object args: encryption handler class constructor arguments.
    :param object kwargs: encryption handler class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           handler with the same name, replace it with
                           the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidEncryptionHandlerTypeError: invalid encryption handler type error.
    :raises InvalidEncryptionHandlerNameError: invalid encryption handler name error.
    :raises DuplicatedEncryptionHandlerError: duplicated encryption handler error.

    :returns: encryption handler class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available encryption handlers.

        :param type cls: encryption handler class.

        :returns: encryption handler class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        encryption_services.register_encryption_handler(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
city manager module.
"""

from pyrin.core.structs import Manager
from pyrin.database.services import get_current_store
from pyrin.core.globals import _

import pyrin.database.paging.services as paging_services

from demo.common.city.exceptions import CityNotFoundError
from demo.common.city.models import CityEntity
from pyrin.database.transaction.contexts import atomic_context, nested_context, \
    subtransaction_context
from pyrin.database.transaction.decorators import atomic, nested, subtransaction


class CityManager(Manager):
    """
    city manager class.
    """

    def get(self, city_id, **options):
        """
        gets the specified city.

        :param int city_id: city id to get its info.

        :raises CityNotFoundError: city not found error.

        :returns: dict(int id,
                       str name,
                       int province_id: province id)

        :rtype: dict
        """

        store = get_current_store()
        city = store.query(CityEntity).get(city_id)

        if city is None:
            raise CityNotFoundError(_('City [{city_id}] not found.'
                                      .format(city_id=city_id)))

        return city.to_dict()

    def clear(self):
        store = get_current_store()
        return store.query(CityEntity).delete()

    def find(self, name, **filters):
        """
        finds cities with given filters.

        :keyword str name: city name.
        :keyword int province_id: province id.

        :returns: list[dict(int id,
                            str name,
                            int province_id: province id)]

        :rtype: list
        """

        clauses = self._make_find_clause(**filters)
        store = get_current_store()
        entities = store.query(CityEntity)\
            .filter(*clauses)\
            .order_by(CityEntity.id)\
            .paginate(**filters).all()

        return entities

    def _make_find_clause(self, **filters):
        """
        makes the required find clauses based on
        given filters and returns the clauses list.

        :keyword str name: city name.
        :keyword int province_id: province id.

        :rtype: list
        """

        clauses = []

        name = filters.get('name', None)
        province_id = filters.get('province_id', None)

        if name is not None:
            clauses.append(CityEntity.name.icontains(name))

        if province_id is not None:
            clauses.append(CityEntity.province_id == province_id)

        return clauses

    def create(self, **options):
        """
        creates a city.

        :param str name: city name.
        :param int province_id: province id.
        """

        try:
            with atomic_context() as store1:
                # store1 = get_current_store()
                city = CityEntity(name='1')
                store1.add(city)

            with atomic_context():
                city2 = CityEntity(name='FUCK')
                store2 = get_current_store()
                store2.add(city2)
                # raise Exception('FUCK')
        except:
            pass

        city3 = CityEntity(name='PARENT')
        store = get_current_store()
        store.add(city3)
        raise Exception('Im Done.')

    # def create(self, name, province_id, **options):
    #     """
    #     creates a city.
    #
    #     :param str name: city name.
    #     :param int province_id: province id.
    #     """
    #
    #     self.a()
    #     self.b()
    #
    #     city3 = CityEntity(name='fake')
    #     store = get_current_store()
    #     store.add(city3)
    #     raise Exception('Im Done.')
    #
    # @atomic
    # def a(self):
    #     store1 = get_current_store()
    #     city = CityEntity(name='1')
    #     store1.add(city)
    #
    # @atomic
    # def b(self):
    #     city2 = CityEntity(name='2')
    #     store2 = get_current_store()
    #     store2.add(city2)

    # def create(self, name, province_id, **options):
    #     """
    #     creates a city.
    #
    #     :param str name: city name.
    #     :param int province_id: province id.
    #     """
    #
    #     store1 = get_current_store()
    #     city = CityEntity(name='1')
    #     store1.add(city)
    #     a = 0
    #     #store1.commit()
    #
    #     store2 = get_current_store()
    #     city2 = CityEntity(name='2')
    #     store2.add(city2)
    #     # #store2.commit()
    #     #
    #     city3 = CityEntity(name='fake')
    #     store3 = get_current_store()
    #     store3.add(city3)
    #     # # raise Exception('Im Done.')



# -*- coding: utf-8 -*-
"""
orm scoping registry containers atomic module.
"""

import threading

import pyrin.security.session.services as session_services

from pyrin.core.structs import Stack
from pyrin.database.orm.scoping.registry.interface import AbstractAtomicContainerBase


class RequestScopedAtomicContainer(AbstractAtomicContainerBase):
    """
    request scoped atomic container class.
    """

    def __init__(self, scopefunc):
        """
        initializes an instance of RequestScopedAtomicContainer.

        :param function scopefunc: a function that returns a hashable
                                   token representing the current scope.
                                   such as, current thread identifier.
        """

        super().__init__()

        self.scopefunc = scopefunc

        # a dictionary containing all atomic sessions that are present.
        # each atomic session corresponds to a request scoped session.
        # each request could contain a 'Stack' of current atomic sessions.
        # each time an atomic session is requested, it gets the last inserted
        # atomic session, this way it provides the ability to use chained
        # atomic sessions. for example multiple methods using '@atomic' decorator
        # in a single call hierarchy.
        # in the form of: {int key_hash: Stack[object] objects}
        self.registry = {}

    def has(self):
        """
        gets a value indicating that an atomic object is present in the current scope.

        :rtype: bool
        """

        if session_services.is_request_context_available() is True:
            return self.scopefunc() in self.registry

        return False

    def set(self, value):
        """
        sets an atomic value for the current scope.

        :param object value: object to be set in registry.
        """

        stack = self._get_stack(True)
        stack.push(value)

    def clear(self, atomic=False):
        """
        clears the current scope's atomic objects, if any.

        if `atomic=True` is provided, it only clears the
        current atomic object if available.

        :param bool atomic: specifies that it must only clear the current atomic
                            object of current scope. otherwise, it clears all atomic
                            objects of current scope. defaults to False if not provided.
        """

        if atomic is True:
            try:
                self._get_stack().dispose()
            except (KeyError, IndexError):
                pass

            if self._get_stack_length() == 0:
                self._clear_all()
        else:
            self._clear_all()

    def _clear_all(self):
        """
        clears all atomic objects of current scope, if any.
        """

        try:
            del self.registry[self.scopefunc()]
        except KeyError:
            pass

    def get(self):
        """
        gets the current atomic object of current scope if available, otherwise returns None.

        :returns: object
        """

        try:
            return self._get_stack().peek()
        except (KeyError, IndexError):
            return None

    def get_all(self):
        """
        gets all atomic objects of current scope if available, otherwise returns an empty list.

        :returns: list[object]
        """

        try:
            return self._get_stack().peek_all()
        except KeyError:
            return []

    def _get_stack(self, create=False):
        """
        gets the stack of current scope if available.

        otherwise it could either return a new stack or raise an error.

        :param bool create: specifies that if there is no stack available,
                            it should return a new stack, otherwise raise
                            an error. defaults to False if not provided.

        :raises KeyError: key error.

        :rtype: Stack
        """

        key = self.scopefunc()
        try:
            return self.registry[key]
        except KeyError as error:
            if create is True:
                return self.registry.setdefault(key, Stack())
            raise error

    def _get_stack_length(self):
        """
        gets the length of items in the stack of current scope if available.

        if there is no stack available, it returns -1.

        :rtype: int
        """

        try:
            return len(self._get_stack())
        except KeyError:
            return -1


class ThreadScopedAtomicContainer(AbstractAtomicContainerBase):
    """
    thread scoped atomic container class.
    """

    def __init__(self):
        """
        initializes an instance of ThreadScopedAtomicContainer.
        """

        super().__init__()

        self.registry = threading.local()

    def has(self):
        """
        gets a value indicating that an atomic object is present in the current scope.

        :rtype: bool
        """

        return hasattr(self.registry, "stack")

    def set(self, value):
        """
        sets an atomic value for the current scope.

        :param object value: object to be set in registry.
        """

        stack = self._get_stack(True)
        stack.push(value)

    def clear(self, atomic=False):
        """
        clears the current scope's atomic objects, if any.

        if `atomic=True` is provided, it only clears the
        current atomic object if available.

        :param bool atomic: specifies that it must only clear the current atomic
                            object of current scope. otherwise, it clears all atomic
                            objects of current scope. defaults to False if not provided.
        """

        if atomic is True:
            try:
                self._get_stack().dispose()
            except (AttributeError, IndexError):
                pass

            if self._get_stack_length() == 0:
                self._clear_all()
        else:
            self._clear_all()

    def _clear_all(self):
        """
        clears all atomic objects of current scope, if any.
        """

        try:
            del self.registry.stack
        except AttributeError:
            pass

    def get(self):
        """
        gets the current atomic object of current scope if available, otherwise returns None.

        :returns: object
        """

        try:
            return self._get_stack().peek()
        except (AttributeError, IndexError):
            return None

    def get_all(self):
        """
        gets all atomic objects of current scope if available, otherwise returns an empty list.

        :returns: list[object]
        """

        try:
            return self._get_stack().peek_all()
        except AttributeError:
            return []

    def _get_stack(self, create=False):
        """
        gets the stack of current scope if available.

        otherwise it could either return a new stack or raise an error.

        :param bool create: specifies that if there is no stack available,
                            it should return a new stack, otherwise raise
                            an error. defaults to False if not provided.

        :raises AttributeError: attribute error.

        :rtype: Stack
        """

        try:
            return self.registry.stack
        except AttributeError as error:
            if create is True:
                self.registry.stack = Stack()
                return self.registry.stack
            raise error

    def _get_stack_length(self):
        """
        gets the length of items in the stack of current scope if available.

        if there is no stack available, it returns -1.

        :rtype: int
        """

        try:
            return len(self._get_stack())
        except AttributeError:
            return -1



# -*- coding: utf-8 -*-
"""
authentication package.
"""

from pyrin.security.authentication import AuthenticationPackage as BaseAuthenticationPackage


class AuthenticationPackage(BaseAuthenticationPackage):
    """
    authentication package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
security api module.
"""

from pyrin.api.router.decorators import api
from pyrin.core.enumerations import HTTPMethodEnum

import demo.security.services as security_services


# @api('/auth/login', methods=HTTPMethodEnum.POST, authenticated=False)
# def login(username, password, **options):
#     """
#     logs in the provided user and gets a valid token.
#
#     :param str username: username.
#     :param str password: password.
#
#     :returns: a valid token.
#     :rtype: str
#     """
#
#     return security_services.login(username, password, **options)



from sqlalchemy import Integer, Unicode

from pyrin.core.globals import SECURE_FALSE, SECURE_TRUE
from pyrin.database.model.base import CoreEntity
from pyrin.database.orm.sql.schema.base import CoreColumn


class PersonEntity(CoreEntity):
    _table = 'person_test'
    r_id = CoreColumn(name='id', type_=Integer, primary_key=True,
                      autoincrement=False, allow_write=False)
    _r_id = CoreColumn(name='_id', type_=Integer, primary_key=True,
                       autoincrement=False)
    rw_name = CoreColumn(name='name', type_=Unicode)
    _rw_name = CoreColumn(name='_name', type_=Unicode)
    w_age = CoreColumn(name='age', type_=Integer, allow_read=False)
    h_code = CoreColumn(name='code', type_=Integer, allow_read=False, allow_write=False)


person = PersonEntity()
person.r_id = 1
person._r_id = 2
person.rw_name = 'mark'
person._rw_name = 'junior'
person.w_age = 25
person._w_age = 30
person.h_code = 500

r_columns = person.readable_columns
w_columns = person.writable_columns
columns = person.all_columns
r_pk = person.readable_primary_key_columns
w_pk = person.writable_primary_key_columns
pk = person.primary_key_columns
z = 0

res = person.to_dict()
res_all = person.to_dict(readable=SECURE_FALSE)
fr = PersonEntity(**res_all)
fr_all = PersonEntity()
fr_all.from_dict(**res_all, writable=SECURE_FALSE)
fr_ALL = PersonEntity(**res_all, populate_all=SECURE_TRUE)
c = PersonEntity.rw_name.fullname
c2 = PersonEntity.r_id.fullname
v = 0


# -*- coding: utf-8 -*-
"""
demo application main package.
"""

from pyrin.application.base import Application


__version__ = '0.2.2'


class DemoApplication(Application):
    """
    demo application class.

    server should create an instance of this class on startup.
    """
    pass



# -*- coding: utf-8 -*-
"""
users manager module.
"""

import pyrin.security.services as security_services

from pyrin.database.services import get_current_store
from pyrin.security.session.services import get_current_user
from pyrin.core.globals import _
from pyrin.security.users.manager import UsersManager as BaseUsersManager

from demo.security.exceptions import UserNotFoundError
from demo.security.models import UserEntity
from demo.security.users.exceptions import InvalidUserIDError


class UsersManager(BaseUsersManager):
    """
    users manager class.
    """

    def _get(self, user_id, **options):
        """
        gets the specified user.

        :param int user_id: user id to get its info.

        :raises UserNotFoundError: user not found error.

        :rtype: UserEntity
        """

        store = get_current_store()
        user = store.query(UserEntity).get(user_id)

        if user is None:
            raise UserNotFoundError(_('User [{user_id}] not found.'.format(user_id=user_id)))

        return user

    def get(self, **options):
        """
        gets the current user info.

        :raises UserNotFoundError: user not found error.

        :rtype: UserEntity
        """

        user_id = get_current_user()
        return self._get(user_id, **options)

    def _exists(self, user_id):
        """
        gets a value indicating that given user existed.

        :param int user_id: user id to check for existence.

        :raises InvalidUserError: invalid user error.

        :rtype: bool
        """

        if user_id is None:
            raise InvalidUserIDError(_('Input user id could not be None.'))

        store = get_current_store()
        count = store.query(UserEntity.id).filter(UserEntity.id == user_id).count()

        return count > 0

    def is_active(self, user, **options):
        """
        gets a value indicating that given user is active.

        :param int user: user to check its active status.

        :raises InvalidUserError: invalid user error.
        :raises UserNotFoundError: user not found error.

        :rtype: bool
        """

        if self._exists(user) is not True:
            raise UserNotFoundError(_('User [{user}] not found.').format(user=user))

        user_info = self._get(user, **options)
        return user_info.is_active

    def create(self, username, password, first_name, last_name, **options):
        """
        creates a new user based on given inputs.

        :param str username: username.
        :param str password: password.
        :param str first_name: first name.
        :param str last_name: last name.
        """

        entity = UserEntity()
        entity.username = username
        entity.password_hash = security_services.get_password_hash(password)
        entity.first_name = first_name
        entity.last_name = last_name
        entity.is_active = True

        store = get_current_store()
        store.add(entity)

    def get_all(self, **options):
        """
        gets a list of all users.

        :returns: list[UserEntity]
        :rtype: list
        """

        store = get_current_store()
        return store.query(UserEntity).all()



# -*- coding: utf-8 -*-
"""
celery cli module.
"""

import pyrin.task_queues.celery.cli.services as celery_cli_services

from pyrin.cli.decorators import cli, cli_group
from pyrin.core.structs import CLI


@cli_group('celery')
class CeleryCLI(CLI):
    """
    celery cli class.

    this class exposes all celery cli commands.
    """

    _execute_service = celery_cli_services.execute

    @cli
    def worker(self, **options):
        """
        create a worker node.

        :keyword int concurrency: number of child processes processing the queue.
                                  the default is the number of cpus available on
                                  your system.

        :keyword str hostname: set custom hostname for worker node.

        :keyword bool beat: also run the `celery beat` periodic task scheduler.
                            please note that there must only be one instance of
                            this service.

        :keyword str | list[str] queues: list of queues to enable for this worker.
        :keyword bool purge: purges all waiting tasks before the daemon is started.

        :keyword str autoscale: enable autoscaling by providing
                                max_concurrency, min_concurrency.

        :keyword str logfile: path to log file. if no logfile is specified, `stderr` is used.

        :keyword str loglevel: logging level, choose between `DEBUG`, `INFO`, `WARNING`,
                               `ERROR`, `CRITICAL` or `FATAL`.

        :keyword str pidfile: optional file used to store the process pid.
                              the program won't start if this file already exists
                              and the pid is still alive.

        :keyword str optimization: optimization profile to be applied.
                                   it could be from: `default` and `fair`.

        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def beat(self, **options):
        """
        start the beat periodic task scheduler.

        :keyword str logfile: path to log file. if no logfile is specified, `stderr` is used.

        :keyword str loglevel: logging level, choose between `DEBUG`, `INFO`, `WARNING`,
                               `ERROR`, `CRITICAL` or `FATAL`.

        :keyword str pidfile: optional file used to store the process pid.
                              the program won't start if this file already exists
                              and the pid is still alive.

        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def result(self, task_id, **options):
        """
        gives the return value for a given task id.

        :param str task_id: id of the task.

        :keyword str task: name of the task (if custom backend).
        :keyword bool traceback: show traceback if any.
        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def inspect(self, method, **options):
        """
        inspect the worker at runtime.

        :param str method: inspect method. it could be from these methods:
                           `active`, `active_queues`, `clock`, `conf`, `memdump`,
                           `memsample`, `objgraph`, `ping`, `query_task`, `registered`,
                           `report`, `reserved`, `revoked`, `scheduled` and `stats`.

        :keyword bool include_defaults: this is only for `conf` method.
        :keyword int samples_count: this is only for `memdump` method.
        :keyword str object_type: this is only for `objgraph` method.
        :keyword int count: this is only for `objgraph` method.
        :keyword int max_depth: this is only for `objgraph` method.
        :keyword str | list[str] task_ids: list of task ids.
                                           this is only for `query_task` method.

        :keyword str | list[str] attributes: this is only for `registered` method.
        :keyword str | list[str] destination: list of destination node names.
        :keyword float timeout: timeout in seconds waiting for reply.
        :keyword bool json: use json as output format.
        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def amqp(self, **options):
        """
        amqp administration shell.

        also works for non-amqp transports (but not ones that
        store declarations in memory).

        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def call(self, task, **options):
        """
        call a task by name.

        :param str task: task name.

        :keyword list args: task positional arguments.
        :keyword dict kwargs: task keyword arguments.
        :keyword str eta: scheduled time in `iso-8601` format.
        :keyword float | int countdown: eta in seconds from now.
        :keyword float | int | str expires: expiry time in seconds or an `iso-8601` date.
        :keyword str serializer: specify serializer to use (default is json).
        :keyword str queue: destination queue name.
        :keyword str exchange: destination exchange name (defaults to the queue exchange).
        :keyword str routing_key: destination routing key (defaults to the queue routing key).

        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def control(self, method, **options):
        """
        workers remote control.

        :param str method: control method. it could be from these methods:
                           `add_consumer`, `autoscale`, `cancel_consumer`,
                           `disable_events`, `election`, `enable_events`,
                           `heartbeat`, `pool_grow`, `pool_restart`,
                           `pool_shrink`, `rate_limit`, `revoke`, `shutdown`,
                           `terminate` and `time_limit`.

        :keyword str queue: the queue name.
                            this is only for `add_consumer`
                            and `cancel_consumer` methods.

        :keyword str exchange: exchange name.
                               this is only for `add_consumer` method.

        :keyword str type: exchange type.
                           this is only for `add_consumer` method.

        :keyword str routing_key: routing key.
                                  this is only for `add_consumer` method.

        :keyword int max_scale: max value for autoscaling.
                                this is only for `autoscale` method.

        :keyword int min_scale: min value for autoscaling.
                                this is only for `autoscale` method.

        :keyword int resize: pool grow or shrink resize value.
                             this is only for `pool_grow` and `pool_shrink` methods.

        :keyword str task: task name.
                           this is only for `rate_limit` and `time_limit` methods.

        :keyword str rate_limit: rate limit value.
                                 for example: 5/s, 5/m or 5/h.
                                 this is only for `rate_limit` method.

        :keyword str | list[str] task_ids: list of task ids.
                                           this is only for `revoke`
                                           and `terminate` methods.

        :keyword str signal: signal name to use for stopping tasks.
                             this is only for `terminate` method.

        :keyword int | float soft_secs: soft time limit for task.
                                        this is only for `time_limit` method.

        :keyword int | float hard_secs: hard time limit for task.
                                        this is only for `time_limit` method.

        :keyword str | list[str] destination: list of destination node names.
        :keyword float timeout: timeout in seconds waiting for reply.
        :keyword bool json: use json as output format.
        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def events(self, **options):
        """
        event-stream utilities.

        :keyword str camera: camera class fully qualified name.
        :keyword float | int frequency: frequency value.
        :keyword bool dump: dump events to stdout.
        :keyword str maxrate: max event streaming rate.
        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def shell(self, **options):
        """
        start shell session with convenient access to celery symbols.

        :keyword bool python: force default python shell.
        :keyword bool ipython: force `ipython` implementation.
        :keyword bool bpython: force `bpython` implementation.
        :keyword bool without_tasks: don't add tasks to locals.
        :keyword bool eventlet: use `eventlet` monkey patches.
        :keyword bool gevent: use `gevent` monkey patches.
        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def list(self, topic, **options):
        """
        get info from broker.

        :param str topic: topic to get info about.
                          for example: `bindings`.

        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def purge(self, **options):
        """
        erase all messages from all known task queues.

        :keyword bool force: don't prompt for verification before deleting messages.
        :keyword str | list[str] queues: list of queue names to purge.
        :keyword str | list[str] exclude_queues: list of queue names not to purge.
        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def logtool(self, action, **options):
        """
        the celery logtool command.

        :param str action: action to be executed. it could be from:
                           `stats`, `traces`, `errors`, `incomplete` and `debug`.

        :keyword str | list[str] files: list of file paths.
        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def status(self, **options):
        """
        show list of worker nodes that are online.

        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def report(self, **options):
        """
        shows information useful to include in bug-reports.

        :keyword bool help: show help for this command.
        """
        pass

    @cli
    def help(self):
        """
        show celery command line help.
        """
        pass



import time

from pyrin.caching.decorators import cached_property, cached_class_property


class Car:
    @cached_property
    def fuel(self):
        return time.time() * 10

    @cached_class_property
    def speed(cls):
        return time.time() * 100


class Plane:
    @cached_property
    def fuel(self):
        return time.time() * 10

    @cached_class_property
    def speed(cls):
        return time.time() * 100


bmw = Car()
f1 = bmw.fuel
s1 = bmw.speed

f2 = bmw.fuel
s2 = bmw.speed


boeing = Plane()
f11 = boeing.fuel
s11 = boeing.speed

f22 = boeing.fuel
s22 = boeing.speed

v = 4


# -*- coding: utf-8 -*-
"""
authentication services module.
"""

from pyrin.application.services import get_component
from pyrin.security.authentication import AuthenticationPackage


def authenticate(client_request, **options):
    """
    authenticates given request and pushes the authenticated data into request context.

    if authentication fails, authenticated data will not be pushed into request context.

    :param CoreRequest client_request: request to be authenticated.

    :raises AuthenticationFailedError: authentication failed error.
    :raises InvalidPayloadDataError: invalid payload data error.
    :raises AccessTokenRequiredError: access token required error.
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME).authenticate(client_request,
                                                                            **options)



# -*- coding: utf-8 -*-
"""
users exceptions module.
"""

from pyrin.core.exceptions import CoreBusinessException, CoreException


class UsersManagerException(CoreException):
    """
    users manager exception.
    """
    pass


class UsersManagerBusinessException(CoreBusinessException,
                                    UsersManagerException):
    """
    users manager business exception.
    """
    pass


class InvalidUserIDError(UsersManagerBusinessException):
    """
    invalid user id error.
    """
    pass



# -*- coding: utf-8 -*-
"""
serializer exceptions module.
"""

from pyrin.core.exceptions import CoreException


class SerializerManagerException(CoreException):
    """
    serializer manager exception.
    """
    pass


class InvalidSerializerTypeError(SerializerManagerException):
    """
    invalid serializer type error.
    """
    pass


class DuplicatedSerializerError(SerializerManagerException):
    """
    duplicated serializer error.
    """
    pass



# -*- coding: utf-8 -*-
"""
caching local handlers exceptions module.
"""

from pyrin.core.exceptions import CoreException


class CachingLocalHandlersException(CoreException):
    """
    caching local handlers exception.
    """
    pass


class InvalidCacheContainerTypeError(CachingLocalHandlersException):
    """
    invalid cache container type error.
    """
    pass


class InvalidCacheItemTypeError(CachingLocalHandlersException):
    """
    invalid cache item type error.
    """
    pass


class CacheClearanceLockTypeIsRequiredError(CachingLocalHandlersException):
    """
    cache clearance lock type is required error.
    """
    pass


class InvalidCacheClearCountError(CachingLocalHandlersException):
    """
    invalid cache clear count error.
    """
    pass


class InvalidChunkSizeError(CachingLocalHandlersException):
    """
    invalid chunk size error.
    """
    pass


class CachePersistentLockTypeIsRequiredError(CachingLocalHandlersException):
    """
    cache persistent lock type is required error.
    """
    pass


class CacheVersionIsRequiredError(CachingLocalHandlersException):
    """
    cache version is required error.
    """
    pass



# -*- coding: utf-8 -*-
"""
template component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.template import TemplatePackage
from pyrin.template.manager import TemplateManager


@component(TemplatePackage.COMPONENT_NAME)
class TemplateComponent(Component, TemplateManager):
    """
    template component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
accounting package.
"""

from pyrin.packaging.base import Package


class AccountingPackage(Package):
    """
    accounting package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'accounting.component'



# -*- coding: utf-8 -*-
"""
encryption handlers exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class EncryptionHandlerException(CoreException):
    """
    encryption handler exception.
    """
    pass


class EncryptionHandlerBusinessException(CoreBusinessException,
                                         EncryptionHandlerException):
    """
    encryption handler business exception.
    """
    pass


class InvalidEncryptedValueError(EncryptionHandlerBusinessException):
    """
    invalid encrypted value error.
    """
    pass


class EncryptionHandlerMismatchError(EncryptionHandlerBusinessException):
    """
    encryption handler mismatch error.
    """
    pass



# -*- coding: utf-8 -*-
"""
packaging exceptions module.
"""

from pyrin.core.exceptions import CoreException


class PackagingManagerException(CoreException):
    """
    packaging manager exception.
    """
    pass


class InvalidPackageNameError(PackagingManagerException):
    """
    invalid package name error.
    """
    pass


class ComponentModuleNotFoundError(PackagingManagerException):
    """
    component module not found error.
    """
    pass


class BothUnitAndIntegrationTestsCouldNotBeLoadedError(PackagingManagerException):
    """
    both unit and integration tests could not be loaded error.
    """
    pass


class InvalidPackagingHookTypeError(PackagingManagerException):
    """
    invalid packaging hook type error.
    """
    pass


class CircularDependencyDetectedError(PackagingManagerException):
    """
    circular dependency detected error.
    """
    pass


class SelfDependencyDetectedError(PackagingManagerException):
    """
    self dependency detected error.
    """
    pass


class SubPackageDependencyDetectedError(PackagingManagerException):
    """
    sub-package dependency detected error.
    """
    pass


class PackageExternalDependencyError(PackagingManagerException):
    """
    package external dependency error.
    """
    pass


class PackageNotExistedError(PackagingManagerException):
    """
    package not existed error.
    """
    pass


class PackageIsIgnoredError(PackagingManagerException):
    """
    package is ignored error.
    """
    pass


class PackageIsDisabledError(PackagingManagerException):
    """
    package is disabled error.
    """
    pass
